# ======================================================================
# Copyright 2016 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

from enum import Enum
from time import ctime
from typing import Any, Callable, Tuple

from supervisor.http import NOT_DONE_YET
from supervisor.xmlrpc import RPCError

# HTML page names
SUPVISORS_PAGE = 'index.html'
CONCILIATION_PAGE = 'conciliation.html'
HOST_INSTANCE_PAGE = 'host_instance.html'
PROC_INSTANCE_PAGE = 'proc_instance.html'
APPLICATION_PAGE = 'application.html'
TAIL_PAGE = 'tail.html'
MAIN_TAIL_PAGE = 'maintail.html'
STDOUT_PAGE = 'logtail/%s'
STDERR_PAGE = 'logtail/%s/stderr'
MAIN_STDOUT_PAGE = 'mainlogtail'

# gravity classes for messages
# use of 'erro' instead of 'error' in order to avoid HTTP error log traces
Info = 'info'
Warn = 'warn'
Error = 'erro'

# Web UI symbols
MASTER_SYMBOL = '\u272A'
SUB_SYMBOL = '\u21B3'
SHEX_SHRINK = '[\u2013]'
SHEX_EXPAND = '[+]'


# entry types in a Process table
class ProcessRowTypes(Enum):
    """ Used to sort entries in a process table. """
    INSTANCE_PROCESS, SUPERVISOR_PROCESS, APPLICATION_PROCESS, APPLICATION = range(4)


def format_gravity_message(message):
    """ Add a gravity to a message if not present."""
    if not isinstance(message, tuple):
        # gravity is not set by Supervisor so let's deduce it
        if 'ERROR' in message:
            message = message.replace('ERROR: ', '')
            gravity = Error
        elif 'unexpected rpc fault' in message:
            gravity = Warn
        else:
            gravity = Info
        return gravity, message
    # in other cases, Supervisor message is suitable
    return message


def print_message(root: Any, gravity: str, message: str,
                  current_time: float, identifier: str):
    """ Print message as a result of action. """
    # print time and source
    root.findmeld('time_mid').content(ctime(current_time))
    root.findmeld('identifier_mid').content(identifier)
    # print message
    elt = root.findmeld('message_mid')
    if message is None:
        # always set the refresh time by default
        elt.attrib['class'] = 'empty'
        elt.content('')
    else:
        elt.attrib['class'] = gravity
        elt.content(message)


def format_message(msg, identifier=None) -> str:
    """ Define a global message structure. """
    return f'{msg} at {ctime()}' + (f' on {identifier}' if identifier else '')


def info_message(msg, identifier=None) -> Tuple[str, str]:
    """ Define an information message. """
    return Info, format_message(msg, identifier)


def warn_message(msg, identifier=None) -> Tuple[str, str]:
    """ Define a warning message. """
    return Warn, format_message(msg, identifier)


def error_message(msg, identifier=None) -> Tuple[str, str]:
    """ Define an error message. """
    return Error, format_message(msg, identifier)


def delayed_message(fct: Callable, msg: str, identifier=None) -> Callable:
    """ Define a delayed message. """
    def on_wait():
        return fct(msg, identifier)

    on_wait.delay = 0.05
    return on_wait


def delayed_info(msg: str, identifier=None) -> Callable:
    """ Define a delayed information message. """
    return delayed_message(info_message, msg, identifier)


def delayed_warn(msg, identifier=None) -> Callable:
    """ Define a delayed warning message. """
    return delayed_message(warn_message, msg, identifier)


def delayed_error(msg, identifier=None) -> Callable:
    """ Define a delayed error message. """
    return delayed_message(error_message, msg, identifier)


def generic_rpc(rpc_intf, rpc_name: str, args: tuple, success_msg: str) -> Callable:
    """ Generic wrapper for an XML-RPC in Supervisor.

    :param rpc_intf: the RPC interface
    :param rpc_name: the RPC name
    :param args: the arguments of the RPC
    :param success_msg: the message in case of success
    :return: a callable for deferred result
    """
    try:
        cb = getattr(rpc_intf, rpc_name)(*args)
    except RPCError as e:
        return delayed_error(f'{rpc_name}: {e.text}')
    # process the result if callable
    if callable(cb):
        def onwait():
            try:
                result = cb()
            except RPCError as exc:
                return error_message(f'{rpc_name}: {exc.text}')
            if result is NOT_DONE_YET:
                return NOT_DONE_YET
            return info_message(success_msg)

        onwait.delay = 0.1
        return onwait
    # process the result if directly available
    return delayed_info(success_msg)


def update_attrib(elt, attribute: str, value: str) -> None:
    """ Add a new item to the attribute of an element.

    :param elt: the element to update.
    :param attribute: the element attribute to update.
    :param value: the value to add.
    :return: None.
    """
    current_value = elt.attrib.get(attribute, '')
    # NOTE: test on the split string because 'on' is in 'button'
    if value not in current_value.split():
        elt.attrib[attribute] = f'{current_value} {value}'.strip()


def apply_shade(elt, shaded: bool) -> None:
    """ Apply shade on element.

    :param elt: the element to shade.
    :param shaded: the shade status.
    :return: None.
    """
    if shaded:
        update_attrib(elt, 'class', 'shaded')
    else:
        update_attrib(elt, 'class', 'brightened')



# ======================================================================
# Copyright 2018 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

import math
from distutils.util import strtobool
from typing import Optional, Tuple
from urllib.parse import quote

from supvisors.process import ProcessStatus
from supvisors.ttypes import StartingStrategies, NameList
from supvisors.utils import get_bit, set_bit
from .webutils import SUPVISORS_PAGE, error_message

# form parameters
SERVER_URL = 'SERVER_URL'
SERVER_PORT = 'SERVER_PORT'
PATH_TRANSLATED = 'PATH_TRANSLATED'

IDENTIFIER = 'ident'  # navigation
APPLI = 'appname'  # navigation

ACTION = 'action'
NAMESPEC = 'namespec'   # used for actions
PROCESS = 'processname'  # used to tail (but also to display statistics)

PERIOD = 'period'
STRATEGY = 'strategy'
CPU = 'cpuid'
NIC = 'nic'
DISK_STATS = 'diskstats'
PARTITION = 'partition'
DEVICE = 'device'
AUTO = 'auto'  # auto-refresh
LIMIT = 'limit'

MESSAGE = 'message'
GRAVITY = 'gravity'

APP_SHRINK_EXPAND = 'ashex'
PROC_SHRINK_EXPAND = 'pshex'

# response parameters
HEADERS = 'headers'
LOCATION = 'Location'


class ViewContext:
    """ Class used to retrieve the parameters selected on the web page.
    It is also used to format href in html pages. """

    def __init__(self, context):
        """ Define attributes for statistics selection. """
        # store the HTML context
        self.http_context = context
        # keep references to Supvisors instance
        self.supvisors = context.supervisord.supvisors
        self.logger = self.supvisors.logger
        # keep reference to the local identifier
        self.local_identifier = self.supvisors.mapper.local_identifier
        # initialize parameters
        self.parameters = {}
        self.store_message = None
        self.redirect = False
        # extract parameters from context
        self.update_strategy()
        self.update_auto_refresh()
        self.update_identifier()
        self.update_application_name()
        self.update_process_name()
        self.update_namespec()
        self.update_application_shrink_expand()
        self.update_process_shrink_expand()
        self.update_period()
        self.update_cpu_id()
        self.update_nic_name()
        self.update_disk_stats_choice()
        self.update_partition_name()
        self.update_device_name()

    # information extracted
    @property
    def strategy(self) -> StartingStrategies:
        return StartingStrategies[self.parameters[STRATEGY]]

    @property
    def auto_refresh(self) -> bool:
        return self.parameters[AUTO]

    @property
    def identifier(self) -> str:
        return self.parameters[IDENTIFIER]

    @property
    def application_name(self) -> str:
        return self.parameters[APPLI]

    @property
    def process_name(self) -> str:
        return self.parameters[PROCESS]

    @process_name.setter
    def process_name(self, proc_name: str) -> None:
        self.parameters[PROCESS] = proc_name

    @property
    def namespec(self) -> str:
        return self.parameters[NAMESPEC]

    @property
    def application_shex(self) -> str:
        return self.parameters[APP_SHRINK_EXPAND]

    @property
    def process_shex(self) -> str:
        return self.parameters[PROC_SHRINK_EXPAND]

    @property
    def period(self) -> float:
        return self.parameters[PERIOD]

    @property
    def cpu_id(self) -> int:
        return self.parameters[CPU]

    @property
    def nic_name(self) -> str:
        return self.parameters[NIC]

    @property
    def device(self) -> str:
        return self.parameters[DEVICE]

    @property
    def partition(self) -> str:
        return self.parameters[PARTITION]

    @property
    def disk_stats(self) -> str:
        return self.parameters[DISK_STATS]

    # simple extraction from context form
    @property
    def action(self) -> str:
        """ Extract action requested in context form. """
        return self.http_context.form.get(ACTION)

    @property
    def message(self) -> str:
        """ Extract message text from context form. """
        return self.http_context.form.get(MESSAGE)

    @property
    def gravity(self) -> str:
        """ Extract message gravity from context form. """
        return self.http_context.form.get(GRAVITY)

    def set_message_default(self, message: str, gravity: str):
        """ Set the message and gravity if there is no existing message. """
        if not self.message:
            self.http_context.form[GRAVITY] = gravity
            self.http_context.form[MESSAGE] = message

    # complex extraction from context form
    def update_period(self) -> None:
        """ Extract period from context. """
        default_value = next(iter(self.supvisors.options.stats_periods))
        self._update_float(PERIOD, self.supvisors.options.stats_periods, default_value)

    def update_strategy(self) -> None:
        """ Extract starting strategy from context. """
        self._update_string(STRATEGY, [x.name for x in StartingStrategies],
                            self.supvisors.options.starting_strategy.name)

    def update_auto_refresh(self) -> None:
        """ Extract auto refresh from context. """
        # assign value found or default
        self._update_boolean(AUTO, False)

    def update_identifier(self) -> None:
        """ Extract identifier from context. """
        # assign value found or default
        self._update_string(IDENTIFIER, self.supvisors.mapper.instances, self.local_identifier)

    def update_application_name(self) -> None:
        """ Extract application name from context. """
        self._update_string(APPLI, list(self.supvisors.context.applications.keys()))

    def update_process_name(self) -> None:
        """ Extract process name from context.
        ApplicationView may select a process unknown to this Supvisors instance. """
        status = self.supvisors.context.instances[self.parameters[IDENTIFIER]]
        # consider processes running on this Supvisors instance + supervisord
        running_processes = [x.namespec for x in status.running_processes()]
        running_processes.append('supervisord')
        self._update_string(PROCESS, running_processes)

    def update_namespec(self) -> None:
        """ Extract namespec from context. """
        value = None
        str_value = self.http_context.form.get(NAMESPEC)
        if str_value:
            # check that value is known to Supvisors
            if self.supvisors.context.is_namespec(str_value):
                value = str_value
            else:
                self.store_message = error_message(f'Incorrect {NAMESPEC}: {str_value}')
        # assign value found or default
        self.logger.trace(f'ViewContext.update_namespec: {NAMESPEC} set to {value}')
        self.parameters[NAMESPEC] = value

    def update_cpu_id(self) -> None:
        """ Extract CPU id from context. """
        self._update_integer(CPU, list(range(self.get_nb_cores(self.local_identifier) + 1)))

    def update_nic_name(self) -> None:
        """ Extract network interface name from context.
        Only the HostInstanceView displays interface data, so it's local. """
        stats_instance = self.get_instance_stats()
        interfaces = stats_instance.net_io.keys() if stats_instance else []
        default_value = next(iter(interfaces), None)
        self._update_string(NIC, interfaces, default_value)

    def update_disk_stats_choice(self) -> None:
        """ Extract the disk stats choice. """
        self._update_string(DISK_STATS, ['usage', 'io'], 'io')

    def update_partition_name(self) -> None:
        """ Extract interface name from context.
        Only the HostInstanceView displays interface data, so it's local. """
        stats_instance = self.get_instance_stats()
        partitions = stats_instance.disk_usage.keys() if stats_instance else []
        default_value = next(iter(partitions), None)
        self._update_string(PARTITION, partitions, default_value)

    def update_device_name(self) -> None:
        """ Extract interface name from context.
        Only the HostInstanceView displays interface data, so it's local. """
        stats_instance = self.get_instance_stats()
        devices = stats_instance.disk_io.keys() if stats_instance else []
        default_value = next(iter(devices), None)
        self._update_string(DEVICE, devices, default_value)

    @staticmethod
    def _get_default_shex(nb_items: int, expanded: bool) -> bytearray:
        """ Get a default shex bytearray filled with 1 if expanded.

        :param nb_items: the number of items in the shex.
        :param expanded: a status telling if the bytearray should be filled with 0 or 1.
        :return: the application shex bytearray.
        """
        base_value = 0xff if expanded else 0
        return bytearray([base_value] * math.ceil(nb_items / 8))

    def get_default_application_shex(self, expanded: bool) -> bytearray:
        """ Get a default application shex bytearray filled with 1 if expanded.

        :param expanded: a status telling if the bytearray should be filled with 0 or 1.
        :return: the application shex bytearray.
        """
        nb_applications = len(self.supvisors.context.applications)
        return ViewContext._get_default_shex(nb_applications, expanded)

    def get_default_process_shex(self, application_name: str, expanded: bool) -> bytearray:
        """ Get a default process shex bytearray filled with 1 if expanded, for the considered application.

        :param application_name: the application name to get the processes from.
        :param expanded: a status telling if the bytearray should be filled with 0 or 1.
        :return: the process shex bytearray.
        """
        nb_processes = len(self.supvisors.context.applications[application_name].processes)
        return ViewContext._get_default_shex(nb_processes, expanded)

    def update_application_shrink_expand(self):
        """ Extract process display choices from context.
        By default, all are displayed. """
        ba = self.get_default_application_shex(True)
        self._update_shex(APP_SHRINK_EXPAND, ba)

    def update_process_shrink_expand(self):
        """ Extract instance process display choices from context.
        By default, all are hidden. """
        if self.application_name:
            ba = self.get_default_process_shex(self.application_name, False)
            self._update_shex(PROC_SHRINK_EXPAND, ba)

    def _update_shex(self, param: str, default_ba: bytearray) -> None:
        """ Extract shex from context. """
        # extract mask from context
        str_value = self.http_context.form.get(param)
        if str_value:
            # check that value has correct format (only hex and size twice the size of the bytearray)
            try:
                value = bytearray.fromhex(str_value)
            except ValueError:
                self.logger.error(f'ViewContext._update_shex: non-hexadecimal {param}')
            else:
                if len(default_ba) != len(value):
                    self.logger.error(f'ViewContext._update_shex: {param} does not fit with the number of items')
                else:
                    default_ba = value
        self.logger.debug(f'ViewContext._update_shex: {param} set to {default_ba.hex()}')
        self.parameters[param] = default_ba.hex()

    def _update_string(self, param: str, check_list: NameList, default_value: str = None):
        """ Extract information from context based on allowed values in check_list. """
        value = default_value
        str_value = self.http_context.form.get(param)
        if str_value:
            # check that value is known to Supvisors
            if str_value in check_list:
                value = str_value
            else:
                self.store_message = error_message(f'Incorrect {param}: {str_value}')
        # assign value found or default
        self.logger.trace(f'ViewContext._update_string: {param} set to {value}')
        self.parameters[param] = value

    def _update_integer(self, param, check_list, default_value=0):
        """ Extract information from context and convert to integer based on allowed values in check_list. """
        value = default_value
        str_value = self.http_context.form.get(param)
        if str_value:
            try:
                int_value = int(str_value)
            except ValueError:
                self.store_message = error_message(f'{param} is not an integer: {str_value}')
            else:
                # check that int_value is defined in check list
                if int_value in check_list:
                    value = int_value
                else:
                    self.store_message = error_message(f'Incorrect {param}: {int_value}')
        # assign value found or default
        self.logger.trace(f'ViewContext._update_integer: {param} set to {value}')
        self.parameters[param] = value

    def _update_float(self, param, check_list, default_value=0.0):
        """ Extract information from context and convert to float based on allowed values in check_list. """
        value = default_value
        str_value = self.http_context.form.get(param)
        if str_value:
            try:
                float_value = float(str_value)
            except ValueError:
                self.store_message = error_message(f'{param} is not a float: {str_value}')
            else:
                # check that float_value is defined in check list
                if float_value in check_list:
                    value = float_value
                else:
                    self.store_message = error_message(f'Incorrect {param}: {float_value}')
        # assign value found or default
        self.logger.trace(f'ViewContext._update_float: {param} set to {value}')
        self.parameters[param] = value

    def _update_boolean(self, param, default_value=False):
        """ Extract information from context and convert to boolean based on allowed values in check_list. """
        value = default_value
        str_value = self.http_context.form.get(param)
        if str_value:
            try:
                value = strtobool(str_value)
            except ValueError:
                self.store_message = error_message(f'{param} is not a boolean-like: {str_value}')
        # assign value found or default
        self.logger.trace(f'ViewContext._update_boolean: {param} set to {value}')
        self.parameters[param] = value

    # URL formatting
    def url_parameters(self, reset_shex, **kwargs):
        """ Return the list of parameters for a URL. """
        parameters = dict(self.parameters, **kwargs)
        if reset_shex:
            del parameters[APP_SHRINK_EXPAND]
        return '&'.join([f'{key}={quote(str(value))}'
                         for key, value in sorted(parameters.items()) if value])

    def format_url(self, identifier: str, page: str, **kwargs):
        """ Format URL from parameters. """
        netloc = ''
        # build network location if identifier is provided
        if identifier:
            instance = self.supvisors.mapper.instances[identifier]
            netloc = f'http://{quote(instance.host_id)}:{instance.http_port}/'
        # shex must be reset if the Supvisors instance changes
        local_identifier = not identifier or identifier == self.local_identifier
        # build URL from netloc, page and attributes
        return f'{netloc}{page}?{self.url_parameters(not local_identifier, **kwargs)}'

    def fire_message(self) -> None:
        """ Set message in context response to be displayed at next refresh. """
        if self.store_message:
            args = {MESSAGE: self.store_message[1], GRAVITY: self.store_message[0]}
            form = self.http_context.form
            # if redirect requested, go back to main page
            path_translated = '/' + SUPVISORS_PAGE if self.redirect else form[PATH_TRANSLATED]
            location = f'{form[SERVER_URL]}{path_translated}?{self.url_parameters(False, **args)}'
            self.http_context.response[HEADERS][LOCATION] = location

    # Statistics
    def get_nb_cores(self, identifier: str) -> int:
        """ Get the number of processors of the host where the Supvisors instance is running. """
        # 2 chances to get the value
        nb_cores = self.supvisors.host_compiler.get_nb_cores(identifier)
        if not nb_cores:
            nb_cores = self.supvisors.process_compiler.get_nb_cores(identifier)
        return nb_cores

    def get_node_characteristics(self):
        """ Get the node characteristics from the stats collector. """
        if self.supvisors.stats_collector:
            return self.supvisors.stats_collector.node_info
        return None

    def get_instance_stats(self, identifier: str = None):
        """ Get the statistics structure related to the identifier and the period selected.
        If no identifier is specified, local identifier is used. """
        stats_identifier = identifier or self.local_identifier
        period = self.parameters.get(PERIOD)
        return self.supvisors.host_compiler.get_stats(stats_identifier, period)

    def get_process_stats(self, namespec: str, identifier: str):
        """ Get the statistics structure related to the process and the period selected.
        The process CPU is updated if SOLARIS mode is expected.
        """
        # return the process statistics for this process
        period = self.parameters.get(PERIOD)
        return self.supvisors.process_compiler.get_stats(namespec, identifier, period)

    def get_process_status(self, namespec: str = None) -> Optional[ProcessStatus]:
        """ Get the ProcessStatus instance related to the process named namespec.
        If none specified, the form namespec is used. """
        namespec = namespec or self.namespec
        if namespec:
            try:
                return self.supvisors.context.get_process(namespec)
            except KeyError:
                self.logger.debug(f'ViewContext.get_process_status: failed to get ProcessStatus from {namespec}')

    # shex access
    def get_application_shex(self, application_name: str) -> Tuple[bool, str]:
        """ Get the expand / shrink value of the application and the shex string to invert it.

        :param application_name: the application name.
        :return: the application shex and the inverted shex.
        """
        ba = bytearray.fromhex(self.application_shex)
        # get the index of the application in context
        idx = sorted(self.supvisors.context.applications).index(application_name)
        # get application shex value
        application_shex = bool(get_bit(ba, idx))
        # get new shex with inverted value for application
        set_bit(ba, idx, not application_shex)
        return application_shex, ba.hex()

    def get_process_shex(self, process_name: str) -> Tuple[bool, str]:
        """ Get the expand / shrink value of the process and the shex string to invert it.

        :param process_name: the process name.
        :return: the process shex and the inverted shex.
        """
        ba = bytearray.fromhex(self.process_shex)
        # get the index of the application in context
        idx = sorted(self.supvisors.context.applications[self.application_name].processes).index(process_name)
        # get application shex value
        process_shex = bool(get_bit(ba, idx))
        # get new shex with inverted value for application
        set_bit(ba, idx, not process_shex)
        return process_shex, ba.hex()

    @staticmethod
    def cpu_id_to_string(idx):
        """ Get a printable form of cpu index. """
        return f'{idx - 1}' if idx > 0 else 'all'



# ======================================================================
# Copyright 2023 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

import select
import struct
import threading
import traceback
from socket import (inet_aton, socket,
                    AF_INET, INADDR_ANY, SOCK_DGRAM, SOL_SOCKET, SO_REUSEADDR,
                    IPPROTO_IP, IPPROTO_UDP, IP_ADD_MEMBERSHIP, IP_MULTICAST_TTL)
from typing import Any, Optional, Set, Tuple

from supervisor.loggers import Logger

from supvisors.ttypes import Ipv4Address, NotificationHeaders, Payload
from .internalinterface import payload_to_bytes, bytes_to_payload
from .mapper import SupvisorsInstanceId

# timeout for polling, in milliseconds
POLL_TIMEOUT = 500

# Default size of the Multicast reception buffer (a lot larger than required, which is less than 200 bytes)
BUFFER_SIZE = 1024


# Emission part
class MulticastSender:
    """ Multicast only TICKS for discovery mode. """

    def __init__(self, mc_group: Ipv4Address, mc_ttl: int, logger: Logger):
        """ Create a socket to multicast messages. """
        self.mc_group: Ipv4Address = mc_group
        self.logger: Logger = logger
        # create the socket
        self.socket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)
        self.socket.setsockopt(IPPROTO_IP, IP_MULTICAST_TTL, mc_ttl)
        logger.info(f'MulticastSender: socket opened to {mc_group}')

    def close(self) -> None:
        """ Close the UDP socket. """
        self.socket.close()

    def send_discovery_event(self, payload: Payload):
        """ Multicast the discovery event.
        It is not necessary to add a message size. """
        self.logger.trace('MulticastSender.emit_message')
        message = payload_to_bytes((NotificationHeaders.DISCOVERY.value, payload))
        try:
            self.socket.sendto(message, self.mc_group)
        except OSError:
            self.logger.error('MulticastSender.emit_message: failed to send DISCOVERY message')
            self.logger.info(f'MulticastSender.emit_message: {traceback.format_exc()}')


# Reception part
class MulticastReceiver(threading.Thread):
    """ The MulticastReceiver receives datagrams, decodes them and put them into the asynchronous queue. """

    def __init__(self, mc_group: Ipv4Address, mc_interface: Optional[str], callback, logger: Logger):
        """ Initialization of the attributes. """
        super().__init__(daemon=True)
        self.logger: Logger = logger
        self.callback = callback
        # save the multicast parameters
        self.mc_group: Ipv4Address = mc_group
        self.mc_interface: Optional[str] = mc_interface
        # the socket will be created when starting the thread
        self.socket: Optional[socket] = None
        # create an event to stop the thread
        self.stop_event: threading.Event = threading.Event()
        # single warning for new clients
        self.log_warnings: Set[Tuple[str, str]] = set()

    def stop(self):
        """ Set a stop event for this thread.

        :return: None
        """
        self.stop_event.set()

    def stopping(self) -> bool:
        """ Access to the loop attribute.

        :return: True if the condition to stop the main loop is reached.
        """
        return self.stop_event.is_set()

    def run(self):
        """ Multicast reception loop. """
        self.logger.info('MulticastReceiver.run: entering Supvisors Discovery main loop')
        self.open_multicast()
        if self.socket:
            # create poller
            poller = select.poll()
            poller.register(self.socket, select.POLLIN)
            # run the loop until stop is called
            while not self.stopping():
                for fd, event in poller.poll(POLL_TIMEOUT):
                    if event == select.POLLIN and fd == self.socket.fileno():
                        datagram = self.socket.recvfrom(BUFFER_SIZE)
                        self.datagram_received(*datagram)
            self.socket.close()
        self.logger.info('MulticastReceiver.run: exiting Supvisors Discovery main loop')

    def datagram_received(self, data, address: Ipv4Address) -> None:
        """ Decode the message received to put it into the asynchronous queue. """
        self.logger.debug(f'MulticastReceiver.datagram_received: size={len(data)} from {address}')
        msg_type, payload = bytes_to_payload(data)
        if self.callback:
            # check the address (log only once)
            updated_address = address[0], payload['http_port']
            if updated_address not in self.log_warnings and address[0] not in payload['ip_addresses']:
                self.logger.warn(f'MulticastReceiver.datagram_received: UDP address={address[0]} does not fit'
                                 f" the discovery event {payload['ip_addresses']}")
                self.log_warnings.add(updated_address)
            # the payload is not needed for now
            self.callback(((payload['identifier'], payload['nick_identifier'], updated_address), (msg_type, ())))

    def open_multicast(self) -> None:
        """ Join the Supvisors Discovery Multicast group. """
        self.logger.debug(f'MulticastReceiver.open_multicast: creating Supvisors Discovery on {self.mc_group}')
        mc_socket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)
        mc_socket.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
        try:
            mc_socket.bind(self.mc_group)
        except (OverflowError, OSError) as exc:
            self.logger.error(f'MulticastReceiver.open_multicast: failed to bind Supvisors Discovery to {self.mc_group}'
                              f' - {str(exc)}')
            self.logger.debug(f'MulticastReceiver.open_multicast: {traceback.format_exc()}')
            return
        self.logger.info(f'MulticastReceiver.open_multicast: Supvisors Discovery opened on {self.mc_group}')
        # set multicast membership according to the configured interface
        try:
            if self.mc_interface:
                mcast_req = struct.pack('=4s4s', inet_aton(self.mc_group[0]), inet_aton(self.mc_interface))
            else:
                mcast_req = struct.pack('=4sl', inet_aton(self.mc_group[0]), INADDR_ANY)
            mc_socket.setsockopt(IPPROTO_IP, IP_ADD_MEMBERSHIP, mcast_req)
        except OSError as exc:
            self.logger.error(f'MulticastReceiver.open_multicast: failed to add membership of Supvisors Discovery'
                              f' on {self.mc_group} to interface={self.mc_interface}'
                              f' - {str(exc)}')
            self.logger.debug(f'MulticastReceiver.open_multicast: {traceback.format_exc()}')
            return
        # assign the socket
        self.socket = mc_socket


class SupvisorsDiscovery:
    """ Class holding the structures used for the emission and the reception of Supvisors discovery events.
    This is activated only if the multicast mode is configured.
    """

    def __init__(self, supvisors: Any) -> None:
        """ Construction of all internal reception blocks.

        :param supvisors: the Supvisors global structure.
        """
        # keep a reference to the Supvisors instance
        self.supvisors = supvisors
        # get the multicast parameters from the Supvisors options
        mc_group = supvisors.options.multicast_group
        mc_ttl = supvisors.options.multicast_ttl
        mc_interface = supvisors.options.multicast_interface
        # create the Multicast message emitter
        local_instance: SupvisorsInstanceId = supvisors.mapper.local_instance
        self.mc_sender: MulticastSender = MulticastSender(mc_group, mc_ttl, supvisors.logger)
        # create the Multicast message receiver
        callback = supvisors.rpc_handler.push_notification
        self.mc_receiver = MulticastReceiver(mc_group, mc_interface, callback, supvisors.logger)
        self.mc_receiver.start()

    def stop(self) -> None:
        """ Stop the Multicast receiver thread. """
        self.mc_sender.close()
        self.mc_receiver.stop()
        self.mc_receiver.join()



# ======================================================================
# Copyright 2024 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

from unittest.mock import call, Mock

import pytest

from supvisors.web.viewmain import *
from .conftest import create_element, to_simple_url


@pytest.fixture
def view(http_context) -> MainView:
    """ Fixture for the instance to test. """
    view = MainView(http_context)
    view.view_ctx = Mock(parameters={}, **{'format_url.side_effect': to_simple_url})
    return view


def test_init(view):
    """ Test the values set at construction. """
    # test action methods storage
    assert sorted(view.global_methods.keys()) == ['sup_restart', 'sup_shutdown']
    assert all(callable(cb) for cb in view.global_methods.values())


def test_write_navigation(mocker, view):
    """ Test the MainView.write_navigation method. """
    mocked_nav = mocker.patch.object(view, 'write_nav')
    mocked_root = create_element()
    view.write_navigation(mocked_root)
    assert mocked_nav.call_args_list == [call(mocked_root, source=view.local_identifier)]


def test_write_status(mocker, supvisors, view):
    """ Test the MainView.write_status method. """
    # patch context
    mocker.patch.object(view.sup_ctx, 'get_state_modes',
                        return_value={'fsm_statecode': SupvisorsStates.DISTRIBUTION.value,
                                      'fsm_statename': SupvisorsStates.DISTRIBUTION.name,
                                      'starting_jobs': True, 'stopping_jobs': False})
    mocker.patch.object(view.sup_ctx, 'conflicting', return_value=False)
    # build root structure
    state_a_mid = create_element()
    starting_mid = create_element()
    stopping_mid = create_element()
    master_mid = create_element()
    mocked_header = create_element({'state_a_mid': state_a_mid, 'master_name_mid': master_mid,
                                    'starting_mid': starting_mid, 'stopping_mid': stopping_mid})
    # test call with no master, starting jobs and not in CONCILIATION
    view.write_status(mocked_header)
    assert mocked_header.findmeld.call_args_list == [call('state_a_mid'), call('starting_mid'), call('stopping_mid'),
                                                     call('master_name_mid')]
    assert state_a_mid.replace.call_args_list == [call('DISTRIBUTION')]
    assert not state_a_mid.attributes.called
    assert not state_a_mid.content.called
    assert state_a_mid.attrib['class'] == ''
    assert starting_mid.attrib['class'] == 'blink'
    assert not starting_mid.replace.called
    assert stopping_mid.attrib['class'] == ''
    assert stopping_mid.replace.call_args_list == [call('')]
    assert master_mid.content.call_args_list == [call('none')]
    mocked_header.reset_all()
    mocker.resetall()
    # test call with master, in CONCILIATION, but without conflicts (expected solved)
    view.sup_ctx.get_state_modes.return_value = {'fsm_statecode': SupvisorsStates.CONCILIATION.value,
                                                 'fsm_statename': SupvisorsStates.CONCILIATION.name,
                                                 'starting_jobs': False, 'stopping_jobs': True}
    supvisors.context.local_status.state_modes.master_identifier = '10.0.0.1:25000'
    view.write_status(mocked_header)
    assert mocked_header.findmeld.call_args_list == [call('state_a_mid'), call('starting_mid'), call('stopping_mid'),
                                                     call('master_name_mid')]
    assert state_a_mid.content.call_args_list == [call('CONCILIATION')]
    assert state_a_mid.attributes.call_args_list == [call(href=CONCILIATION_PAGE)]
    assert not state_a_mid.replace.called
    assert state_a_mid.attrib['class'] == ''
    assert starting_mid.attrib['class'] == ''
    assert starting_mid.replace.call_args_list == [call('')]
    assert stopping_mid.attrib['class'] == 'blink'
    assert not stopping_mid.replace.called
    assert master_mid.content.call_args_list == [call('10.0.0.1')]
    mocked_header.reset_all()
    mocker.resetall()
    # test call with master, in CONCILIATION, and with conflicts
    view.sup_ctx.conflicting.return_value = True
    view.write_status(mocked_header)
    assert mocked_header.findmeld.call_args_list == [call('state_a_mid'), call('starting_mid'), call('stopping_mid'),
                                                     call('master_name_mid')]
    assert state_a_mid.content.call_args_list == [call('CONCILIATION >>')]
    assert state_a_mid.attributes.call_args_list == [call(href=CONCILIATION_PAGE)]
    assert not state_a_mid.replace.called
    assert state_a_mid.attrib['class'] == 'on blink'
    assert starting_mid.attrib['class'] == ''
    assert starting_mid.replace.call_args_list == [call('')]
    assert stopping_mid.attrib['class'] == 'blink'
    assert not stopping_mid.replace.called
    assert master_mid.content.call_args_list == [call('10.0.0.1')]


def test_write_actions(mocker, view):
    """ Test the MainView.write_actions method. """
    mocked_super = mocker.patch('supvisors.web.viewhandler.ViewHandler.write_actions')
    # set context (meant to be set through render)
    view.view_ctx = Mock(**{'format_url.return_value': 'an url'})
    # build root structure
    restart_mid = create_element()
    shutdown_mid = create_element()
    mocked_header = create_element({'restart_a_mid': restart_mid,
                                    'shutdown_a_mid': shutdown_mid})
    # test call
    view.write_actions(mocked_header)
    assert mocked_super.call_args_list == [call(mocked_header)]
    assert mocked_header.findmeld.call_args_list == [call('restart_a_mid'), call('shutdown_a_mid')]
    assert view.view_ctx.format_url.call_args_list == [call('', SUPVISORS_PAGE, **{ACTION: 'sup_restart'}),
                                                       call('', SUPVISORS_PAGE, **{ACTION: 'sup_shutdown'})]
    assert restart_mid.attributes.call_args_list == [call(href='an url')]
    assert shutdown_mid.attributes.call_args_list == [call(href='an url')]


def test_make_callback(view):
    """ Test the MainView.make_callback method. """
    for action_name in list(view.global_methods.keys()):
        view.global_methods[action_name] = Mock(return_value='%s called' % action_name)
    # patch context
    view.view_ctx = Mock(identifier='10.0.0.2')
    # test unknown action
    assert view.make_callback(None, 'dummy') is None
    # test global actions
    for action in ['sup_restart', 'sup_shutdown']:
        assert view.make_callback('', action) == '%s called' % action
        assert view.global_methods[action].call_args_list == [call()]
        view.global_methods[action].reset_mock()


def test_sup_restart_action(mocker, view):
    """ Test the MainView.sup_shutdown_action method. """
    mocked_methods = [mocker.patch('supvisors.web.viewmain.delayed_error', return_value='delayed error'),
                      mocker.patch('supvisors.web.viewmain.delayed_warn', return_value='delayed warning'),
                      mocker.patch('supvisors.web.viewmain.error_message', return_value='error'),
                      mocker.patch('supvisors.web.viewmain.warn_message', return_value='warning')]
    _check_sup_action(mocker, view, view.sup_restart_action, 'restart', *mocked_methods)


def test_sup_shutdown_action(mocker, view):
    """ Test the MainView.sup_shutdown_action method. """
    mocked_methods = [mocker.patch('supvisors.web.viewmain.delayed_error', return_value='delayed error'),
                      mocker.patch('supvisors.web.viewmain.delayed_warn', return_value='delayed warning'),
                      mocker.patch('supvisors.web.viewmain.error_message', return_value='error'),
                      mocker.patch('supvisors.web.viewmain.warn_message', return_value='warning')]
    _check_sup_action(mocker, view, view.sup_shutdown_action, 'shutdown', *mocked_methods)


def _check_sup_action(mocker, view, method_cb, rpc_name, mocked_derror, mocked_dwarn, mocked_error, mocked_warn):
    """ Test the MainView.sup_restart_action & sup_shutdown_action methods. """
    # test RPC error
    mocker.patch.object(view.supvisors.supervisor_data.supvisors_rpc_interface, rpc_name,
                        side_effect=RPCError('failed RPC'))
    assert method_cb() == 'delayed error'
    assert mocked_derror.called
    assert not mocked_dwarn.called
    assert not mocked_error.called
    assert not mocked_warn.called
    # reset mocks
    mocked_derror.reset_mock()
    # test direct result
    mocker.patch.object(view.supvisors.supervisor_data.supvisors_rpc_interface, rpc_name,
                        return_value='not callable object')
    assert method_cb() == 'delayed warning'
    assert not mocked_derror.called
    assert mocked_dwarn.called
    assert not mocked_error.called
    assert not mocked_warn.called



# ======================================================================
# Copyright 2016 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

import re
import time
from typing import Any, Dict, List, Optional, Set, Tuple

from supervisor.loggers import Logger
from supervisor.options import make_namespec, split_namespec

from supvisors.external_com.eventinterface import EventPublisherInterface
from .application import ApplicationRules, ApplicationStatus
from .instancestatus import SupvisorsInstanceStatus
from .process import ProcessRules, ProcessStatus
from .ttypes import (ApplicationStates, SupvisorsInstanceStates, SupvisorsStates,
                     WORKING_STATES, CLOSING_STATES,
                     Ipv4Address, NameList, Payload, PayloadList, LoadMap)


class Context:
    """ The Context class holds the main data of Supvisors.

    Attributes are:
        - supvisors: the Supvisors global structure ;
        - instances: the dictionary of all SupvisorsInstanceStatus (key is Supvisors identifier) ;
        - applications: the dictionary of all ApplicationStatus (key is application name) ;
        - start_date: the date since Supvisors entered the INITIALIZATION state ;
        - last_state_modes: the last Supvisors State and Modes published.
    """

    # annotation types
    InstancesMap = Dict[str, SupvisorsInstanceStatus]
    ApplicationsMap = Dict[str, ApplicationStatus]

    # attributes
    supvisors: Any = None
    start_date: float = 0.0
    instances: InstancesMap = None
    applications: ApplicationsMap = None
    last_state_modes: Payload = None

    def __init__(self, supvisors: Any):
        """ Initialization of the attributes. """
        # keep a reference of the Supvisors data
        self.supvisors = supvisors
        # the Supvisors instances declared statically
        self.instances = {identifier: SupvisorsInstanceStatus(supvisors_id, supvisors)
                          for identifier, supvisors_id in self.supvisors.mapper.instances.items()}
        # the applications known to Supvisors
        self.applications = {}

    def reset(self) -> None:
        """ Reset the context to prepare a new synchronization phase.

        :return: None
        """
        self.local_status.state_modes.master_identifier = ''
        self.start_date = time.monotonic()
        for status in self.instances.values():
            status.reset()

    @property
    def logger(self) -> Logger:
        """ Get the Supvisors logger. """
        return self.supvisors.logger

    @property
    def external_publisher(self) -> Optional[EventPublisherInterface]:
        """ Get the Supvisors external publisher. """
        return self.supvisors.external_publisher

    @property
    def local_identifier(self) -> str:
        """ Get the local Supvisors instance identifier. """
        return self.supvisors.mapper.local_identifier

    @property
    def local_status(self) -> SupvisorsInstanceStatus:
        """ Get the local Supvisors instance structure. """
        return self.instances[self.local_identifier]

    @property
    def local_sequence_counter(self) -> int:
        """ Get last local TICK sequence counter, used for node invalidation. """
        return self.local_status.sequence_counter

    # Master operations
    @property
    def master_identifier(self) -> str:
        """ Get the identifier of the Supvisors Master instance. """
        return self.local_status.state_modes.master_identifier

    @master_identifier.setter
    def master_identifier(self, identifier: str) -> None:
        """ Set the identifier of the known Supvisors Master instance. """
        self.logger.info(f'Context.master_identifier: {identifier}')
        self.publish_state_modes({'master_identifier': identifier})

    @property
    def is_master(self) -> bool:
        """ Return True if the local Supvisors instance is the Supvisors Master instance. """
        return self.master_identifier == self.local_identifier

    @property
    def master_instance(self) -> Optional[SupvisorsInstanceStatus]:
        """ Get local Supvisors instance structure. """
        return self.instances.get(self.master_identifier)

    @property
    def supvisors_state(self) -> Optional[SupvisorsStates]:
        """ Get the supvisors state of the Supvisors Master instance. """
        if self.master_instance:
            return self.master_instance.state_modes.state
        return None

    def elect_master(self, running_identifiers: Optional[NameList] = None) -> None:
        """ Select the Master Supvisors instance among the possible candidates. """
        if not running_identifiers:
            running_identifiers = self.running_identifiers()
        self.logger.info(f'Context.elect_master: Supvisors Master instance election among {running_identifiers}')
        if running_identifiers:
            # elect master instance among working instances only if not fixed before
            # of course, master instance must be running
            self.logger.debug(f'Context.elect_master: master_identifier={self.master_identifier}')
            if not self.master_identifier or self.master_identifier not in running_identifiers:
                # choose Master among the core instances because they are expected to be more stable
                #   this logic is kept independently of CORE being selected as synchro_options
                core_identifiers = self.supvisors.mapper.core_identifiers
                self.logger.debug(f'Context.elect_master: core_identifiers={core_identifiers}')
                if core_identifiers:
                    running_core_identifiers = set(running_identifiers).intersection(core_identifiers)
                    if running_core_identifiers:
                        running_identifiers = running_core_identifiers
                # arbitrarily choice: master instance has the 'lowest' nick_identifier among running instances
                self.master_identifier = min(running_identifiers,
                                             key=lambda x: self.supvisors.mapper.instances[x].nick_identifier)

    # States and Modes
    def publish_state_modes(self, event: Payload) -> None:
        """ Publish the Supvisors instance state and modes.
        FIXME: it's not only publication but update too

        This information is provided internally by the local Supvisors instance, and it will be published to all
        connected Supvisors instances.

        :param event: the state or the mode updated.
        :return: None.
        """
        # keys are Starter, Stopper, fsm_state
        # internal update due to FSM state change or Starter / Stopper progress change
        changed, state_modes = self.local_status.apply_state_modes(event)
        if changed:
            # on change, publish the local Supvisors state and modes to the other Supvisors instances
            self.supvisors.rpc_handler.send_state_event(state_modes.serial())
            # publish SupvisorsInstanceStatus and SupvisorsStatus
            self.export_status(self.local_status)

    def get_state_modes(self) -> Payload:
        """ Get the Supvisors state and modes, based on all connected Supvisors instances.
        Supvisors state is the FSM state and is a reflection of the Supvisors Master instance state.
        Supvisors modes consist in the existence of starting or stopping jobs across all connected Supvisors instances.

        :return: the Supvisors state and the identifiers of the Supvisors instances having starting or stopping jobs
        """
        payload = self.local_status.state_modes.serial()
        # overwrite starting_jobs and starting_jobs based on a synthesis from all Supvisors instances
        starting, stopping = [], []
        for identifier, status in self.instances.items():
            if status.state_modes.starting_jobs:
                starting.append(identifier)
            if status.state_modes.stopping_jobs:
                stopping.append(identifier)
        # update the payload
        # NOTE: discovery mode is only based on the local Supvisors instance
        #       there's no chance that Supvisors instances in discovery mode would communicate with Supvisors instances
        #       that are NOT in discovery mode
        payload.update({'starting_jobs': starting, 'stopping_jobs': stopping})
        return payload

    def _publish_state_mode(self) -> None:
        """ Publish the new state and modes if it differs from the latest publication.

        :return: None
        """
        current_state_modes = self.get_state_modes()
        if self.last_state_modes != current_state_modes:
            self.last_state_modes = current_state_modes
            self.external_publisher.send_supvisors_status(current_state_modes)

    def export_status(self, status: SupvisorsInstanceStatus) -> None:
        """ Publish SupvisorsInstanceStatus and SupvisorsStatus to Supvisors listeners. """
        if self.external_publisher:
            self.external_publisher.send_instance_status(status.serial())
            self._publish_state_mode()

    # methods on nodes
    def is_valid(self, identifier: str, nick_identifier: str,
                 ipv4_address: Ipv4Address) -> Optional[SupvisorsInstanceStatus]:
        """ Check the validity of the message emitter.
        Validity is ok if the identifier is known with the correct IP address and not declared isolated. """
        ip_address, http_port = ipv4_address
        identifiers = self.supvisors.mapper.filter([identifier, nick_identifier])
        if len(identifiers) != 1:
            # multiple resolution not expected here
            return None
        status = self.instances[identifiers[0]]
        if (not status.isolated and ip_address in status.supvisors_id.ip_addresses
                and status.supvisors_id.http_port == http_port):
            return status
        return None

    def get_nodes_load(self) -> LoadMap:
        """ Get the Supvisors instances load grouped by node.

        :return: The nodes load
        """
        return {ip_address: sum(self.instances[identifier].get_load()
                                for identifier in identifiers)
                for ip_address, identifiers in self.supvisors.mapper.nodes.items()}

    # methods on instances
    def initial_running(self) -> bool:
        """ Return True if all Supervisor instances are in RUNNING state. """
        return all(status.state == SupvisorsInstanceStates.RUNNING
                   for identifier, status in self.instances.items()
                   if identifier in self.supvisors.mapper.initial_identifiers)

    def all_running(self) -> bool:
        """ Return True if all Supervisor instances are in RUNNING state. """
        return all(status.state == SupvisorsInstanceStates.RUNNING
                   for status in self.instances.values())

    def running_identifiers(self) -> NameList:
        """ Return the identifiers of the Supervisor instances in RUNNING state. """
        return self.instances_by_states([SupvisorsInstanceStates.RUNNING])

    def running_core_identifiers(self) -> bool:
        """ Check if core SupvisorsInstanceStatus are in RUNNING state.

        :return: True if all core SupvisorsInstanceStatus are in RUNNING state
        """
        core_identifiers = self.supvisors.mapper.core_identifiers
        if core_identifiers:
            return all(status.state == SupvisorsInstanceStates.RUNNING
                       for identifier, status in self.instances.items()
                       if identifier in core_identifiers)
        return False

    def isolated_instances(self) -> NameList:
        """ Return the identifiers of the Supervisors in ISOLATED state. """
        return self.instances_by_states([SupvisorsInstanceStates.ISOLATED])

    def valid_instances(self) -> NameList:
        """ Return the identifiers of the Supervisors NOT in ISOLATED state. """
        return self.instances_by_states([SupvisorsInstanceStates.UNKNOWN,
                                         SupvisorsInstanceStates.CHECKING, SupvisorsInstanceStates.CHECKED,
                                         SupvisorsInstanceStates.RUNNING, SupvisorsInstanceStates.SILENT])

    def instances_by_states(self, states: List[SupvisorsInstanceStates]) -> NameList:
        """ Return the Supervisor identifiers sorted by Supervisor state. """
        return [identifier for identifier, status in self.instances.items() if status.state in states]

    def activate_checked(self):
        """ Once authorized, a Supvisors instance will be set to RUNNING only when it is certain it won't interfere
        with any sequencing in progress. """
        for status in self.instances.values():
            if status.state == SupvisorsInstanceStates.CHECKED:
                status.state = SupvisorsInstanceStates.RUNNING

    def invalid_unknown(self):
        """ This can be triggered by the FSM when TIMEOUT is set in synchro_options.
        After the synchro_timeout has passed, all UNKNOWN Supvisors instances are invalidated. """
        for status in self.instances.values():
            if status.state == SupvisorsInstanceStates.UNKNOWN:
                # nothing to do on processes as none received yet
                self.invalid(status)

    def invalid(self, status: SupvisorsInstanceStatus, fence=None) -> None:
        """ Declare SILENT or ISOLATED the SupvisorsInstanceStatus in parameter, according to the auto_fence option.

        The local Supvisors instance is never set to ISOLATED, whatever the option is set or not.
        Always give it a chance to restart.

        :param status: the Supvisors instance to invalid.
        :param fence: True when the remote Supvisors instance has isolated the local Supvisors instance.
        """
        if status.identifier == self.local_identifier:
            # A very few events can cause this situation:
            # 1. a network failure (XML-RPC request)
            # 2. a discrepancy has been detected between the internal context and the process events received
            #    a new CHECKING phase is required
            # NOTE: the local Supvisors instance cannot be ISOLATED from itself
            self.logger.critical('Context.invalid: local Supvisors instance is either SILENT or inconsistent')
            status.state = SupvisorsInstanceStates.SILENT
        elif fence or (self.supvisors.options.auto_fence and self.supvisors_state in WORKING_STATES):
            # isolation of the remote Supvisors instance can be initiated when:
            #   - the remote Supvisors instance has isolated the local Supvisors instance (auth exchange)
            #   - the remote Supvisors instance has become non-responsive, and the option auto_fence is activated,
            #     the Supvisors FSM is in WORKING_STATES.
            status.state = SupvisorsInstanceStates.ISOLATED
        else:
            status.state = SupvisorsInstanceStates.SILENT
        # publish SupvisorsInstanceStatus and SupvisorsStatus
        self.export_status(status)

    # methods on applications / processes
    def get_managed_applications(self) -> Dict[str, ApplicationStatus]:
        """ Get the managed applications (as defined in rules file).

        :return: the managed applications.
        """
        return {application_name: application for application_name, application in self.applications.items()
                if application.rules.managed}

    def is_namespec(self, namespec: str) -> bool:
        """ Check if the namespec is valid with the context.

        :return: True if the namespec is valid.
        """
        application_name, process_name = split_namespec(namespec)
        if application_name not in self.applications:
            return False
        if process_name:
            application_status = self.applications[application_name]
            return process_name in application_status.processes
        return True

    def get_process(self, namespec: str) -> Optional[ProcessStatus]:
        """ Return the ProcessStatus corresponding to the namespec.

        :param namespec: the process namespec.
        :return: the corresponding ProcessStatus.
        """
        application_name, process_name = split_namespec(namespec)
        return self.applications[application_name].processes[process_name]

    def find_runnable_processes(self, regex: str) -> List[ProcessStatus]:
        """ Get all processes whose namespec matches the regex.
        The processes shall not be already running.

        :return: the candidate processes.
        """
        return [process for application in self.applications.values()
                for process in application.processes.values()
                if re.search(rf'{regex}', process.namespec) and not process.running()]

    def conflicting(self) -> bool:
        """ Check if any conflicting ProcessStatus is detected.

        :return: True if at least one conflict is detected
        """
        return any((process.conflicting() for application in self.applications.values()
                    for process in application.processes.values()
                    if application.rules.managed))

    def conflicts(self) -> List[ProcessStatus]:
        """ Get all conflicting processes.

        :return: the list of conflicting ProcessStatus
        """
        return [process for application in self.applications.values()
                for process in application.processes.values()
                if application.rules.managed and process.conflicting()]

    def setdefault_application(self, application_name: str) -> Optional[ApplicationStatus]:
        """ Return the application corresponding to application_name if found,
        otherwise load rules from the rules file, create a new application entry if rules exist and return it.
        Applications that are not defined in the rules files will not be stored in the Supvisors context.

        :param application_name: the name of the application.
        :return: the application stored in the Supvisors context.
        """
        # find existing application
        application = self.applications.get(application_name)
        if not application:
            # load rules from rules file
            rules = ApplicationRules(self.supvisors)
            if self.supvisors.parser:
                # apply default starting strategy from options
                rules.starting_strategy = self.supvisors.options.starting_strategy
                self.supvisors.parser.load_application_rules(application_name, rules)
                self.logger.debug(f'Context.setdefault_application: application={application_name} rules={rules}')
            # create new instance
            application = ApplicationStatus(application_name, rules, self.supvisors)
            self.applications[application_name] = application
        return application

    def setdefault_process(self, identifier: str, info: Payload) -> Optional[ProcessStatus]:
        """ Return the process corresponding to info if found,
        otherwise load rules from the rules file, create a new process entry if rules exist and return it.
        Processes that are not defined in the rules files will not be stored in the Supvisors context.

        :param identifier: the identification of the Supvisors instance that sent this payload.
        :param info: the payload representing the process.
        :return: the process stored in the Supvisors context.
        """
        application_name, process_name = info['group'], info['name']
        namespec = make_namespec(application_name, info['name'])
        # get application
        application = self.setdefault_application(application_name)
        # search for existing process in application
        process = application.processes.get(process_name)
        new_process = process is None
        if new_process:
            # create process rules
            # by default, apply application starting / running failure strategies
            rules = ProcessRules(self.supvisors)
            rules.starting_failure_strategy = application.rules.starting_failure_strategy
            rules.running_failure_strategy = application.rules.running_failure_strategy
            if self.supvisors.parser:
                # load process rules from rules files
                self.supvisors.parser.load_program_rules(namespec, rules)
                self.logger.debug(f'Context.setdefault_process: namespec={namespec} rules={rules}')
            # create a new ProcessStatus
            process = ProcessStatus(application_name, info['name'], rules, self.supvisors)
        # store the payload in the ProcessStatus
        process.add_info(identifier, info)
        # add a new ProcessStatus to the ApplicationStatus
        if new_process:
            application.add_process(process)
        return process

    def load_processes(self, status: SupvisorsInstanceStatus, all_info: Optional[PayloadList]) -> None:
        """ Load application dictionary from the process information received from the remote Supvisors.

        :param status: the Supvisors instance.
        :param all_info: the process information got from the node.
        :return: None.
        """
        self.logger.trace(f'Context.load_processes: identifier={status.usage_identifier} all_info={all_info}')
        # get SupvisorsInstanceStatus corresponding to identifier
        if all_info is None:
            # the check call in SupervisorProxy failed
            # the remote Supvisors instance is likely starting, restarting or shutting down so defer
            self.logger.warn('Context.load_processes: failed to get all process info from'
                             f' Supvisors={status.usage_identifier}')
            # go back to UNKNOWN to give it a chance at next TICK
            status.state = SupvisorsInstanceStates.UNKNOWN
        else:
            # store processes into their application entry
            for info in all_info:
                # get or create process
                process = self.setdefault_process(status.identifier, info)
                if process:
                    # share the instance to the Supervisor instance that holds it
                    status.add_process(process)
            # re-evaluate application sequences and status
            for application in self.applications.values():
                application.update_sequences()
                application.update()

    # methods on events
    def on_instance_state_event(self, status: SupvisorsInstanceStatus, event: Payload) -> None:
        """ Update the Supvisors instance state and modes.

        :param status: the Supvisors instance that sent the event.
        :param event: the new state and modes.
        :return: None.
        """
        # ISOLATED instances are not updated anymore
        # should not happen as the subscriber should have been disconnected but there may be a tick in the pipe
        if not status.isolated:
            # update the Supvisors instance StateModes status
            status.update_state_modes(event)
            # check if a Master is known to this Supvisors instance and compare with the local's
            # this is considered only if the remote Supvisors instance is not about to restart or shut down
            remote_master = status.state_modes.master_identifier
            if remote_master and status.state_modes.state not in CLOSING_STATES:
                if not self.master_identifier:
                    # the local Supvisors instance doesn't know about a master yet but remote Supvisors instance does
                    # it typically happens when the local Supervisor instance has just been started whereas a Supvisors
                    # group was already operating, so accept the remote perception
                    self.logger.warn(f'Context.on_instance_state_event: accept Master={remote_master}'
                                     f' declared by Supvisors={status.usage_identifier}')
                    self.master_identifier = remote_master
                elif remote_master != self.master_identifier:
                    # ALERT: 2 different perceptions of the master, likely due to a split-brain situation.
                    self.logger.warn('Context.on_instance_state_event: Master conflict. '
                                     f' Local declares Master={self.master_identifier}'
                                     f' - Supvisors={status.usage_identifier} declares Master={remote_master}')
                    # WARN: resetting the Master at local level isn't enough because there may be multiple Supvisors
                    #   instances having different perceptions, and they are still publishing at their own pace.
                    #   So in order to avoid infinite Master conflict, adjudication is made using the same principle
                    #   as the Master selection. This is expected to converge more quickly.
                    candidates = [remote_master, self.master_identifier]
                    self.master_identifier = ''
                    self.elect_master(candidates)
            # publish the new Instance status and Supvisors synthesis
            self.export_status(status)

    def on_authorization(self, status: SupvisorsInstanceStatus, authorized: Optional[bool]) -> bool:
        """ Method called upon reception of an authorization event telling if the remote Supvisors instance
        authorizes the local Supvisors instance to process its events.

        :param status: the Supvisors instance that sent the event.
        :param authorized: the Supvisors instance authorization status.
        :return: True if authorized both ways.
        """
        # check Supvisors instance state
        if status.state != SupvisorsInstanceStates.CHECKING:
            self.logger.error('Context.on_authorization: auth rejected from non-CHECKING'
                              f' Supvisors={status.usage_identifier}')
            return False
        # process authorization status
        if authorized is None:
            # the check call in SupervisorProxy failed
            # the remote Supvisors instance is likely starting, restarting or shutting down so defer
            self.logger.warn('Context.on_authorization: failed to get auth status'
                             f' from Supvisors={status.usage_identifier}')
            # go back to UNKNOWN to give it a chance at next TICK
            status.state = SupvisorsInstanceStates.UNKNOWN
        elif not authorized:
            self.logger.warn('Context.on_authorization: the local Supvisors instance is isolated'
                             f' by Supvisors={status.usage_identifier}')
            self.invalid(status, True)
        else:
            self.logger.info(f'Context.on_authorization: local Supvisors instance is authorized to work with'
                             f' Supvisors={status.usage_identifier}')
            status.state = SupvisorsInstanceStates.CHECKED
            return True
        return False

    def on_local_tick_event(self, event: Payload) -> None:
        """ Method called upon reception of a tick event from the local Supvisors instance.
        The method updates the times of the corresponding SupvisorsInstanceStatus and its ProcessStatus.
        Finally, the updated SupvisorsInstanceStatus is published.

        :param event: the TICK event sent
        :return: None
        """
        # for local Supvisors instance, repeat local data
        counter = event['sequence_counter']
        tick_mtime = event['when_monotonic']
        tick_time = event['when']
        self.local_status.update_tick(counter, tick_mtime, tick_time)
        # trigger hand-shake on first TICK received
        if self.local_status.state in [SupvisorsInstanceStates.UNKNOWN, SupvisorsInstanceStates.SILENT]:
            self.local_status.state = SupvisorsInstanceStates.CHECKING
            self.supvisors.rpc_handler.send_check_instance(self.local_identifier)
        # publish new Supvisors Instance status
        self.export_status(self.local_status)

    def on_tick_event(self, status: SupvisorsInstanceStatus, event: Payload) -> None:
        """ Method called upon reception of a tick event from the remote Supvisors instance, telling that it is active.
        Supvisors checks that the handling of the event is valid in case of auto fencing.
        The method also updates the times of the corresponding SupvisorsInstanceStatus and its ProcessStatus.
        Finally, the updated SupvisorsInstanceStatus is published.
        It is assumed that identifier validity has been checked before.

        :param status: the Supvisors instance from which the event has been received.
        :param event: the TICK event sent.
        :return: None.
        """
        # check if local tick has been received yet
        # NOTE: it is needed because remote ticks are tagged against last local tick received
        if self.local_status.state not in [SupvisorsInstanceStates.CHECKED, SupvisorsInstanceStates.RUNNING]:
            self.logger.debug('Context.on_tick_event: waiting for local tick first')
            return
        # ISOLATED instances are not updated anymore
        if not status.isolated:
            # update the Supvisors instance with the TICK event
            self.supvisors.mapper.assign_stereotypes(status.identifier, event['stereotypes'])
            # for remote Supvisors instance, use local Supvisors instance data
            status.update_tick(event['sequence_counter'], event['when_monotonic'], event['when'],
                               self.local_sequence_counter)
            # trigger hand-shake on first TICK received
            if status.state in [SupvisorsInstanceStates.UNKNOWN, SupvisorsInstanceStates.SILENT]:
                status.state = SupvisorsInstanceStates.CHECKING
                self.supvisors.rpc_handler.send_check_instance(status.identifier)
            # publish new Supvisors Instance status
            self.export_status(status)

    def on_discovery_event(self, identifier: str, nick_identifier: str) -> bool:
        """ Insert a new Supvisors instance if Supvisors is in discovery mode and the origin is unknown.
        If this event is received, the discovery mode is enabled.

        :param identifier: the remote Supvisors identifier.
        :param nick_identifier: the remote Supervisor identifier.
        :return: True if a new Supvisors instance has been inserted.
        """
        if self.supvisors.mapper.check_candidate(identifier, nick_identifier):
            # NOTE: use the first IP address in the list
            item = f'<{nick_identifier}>{identifier}'
            supvisors_id = self.supvisors.mapper.add_instance(item)
            self.logger.info(f'SupvisorsMapper.on_discovery_event: new SupvisorsInstanceId={supvisors_id}')
            self.instances[supvisors_id.identifier] = SupvisorsInstanceStatus(supvisors_id, self.supvisors)
            return True
        return False

    def on_timer_event(self, event: Payload) -> Tuple[NameList, Set[ProcessStatus]]:
        """ Check that all Supvisors instances are still publishing.
        Supvisors considers that there a Supvisors instance is not active if no tick received in last 10s.

        :param event: the timer event.
        :return: the identifiers of the invalidated Supvisors instances and the processes in failure.
        """
        invalidated_identifiers: List[str] = []
        failed_processes: Set[ProcessStatus] = set()
        # use the local TICK counter received as a reference
        sequence_counter = event['sequence_counter']
        # check all Supvisors instances
        for status in self.instances.values():
            if status.is_inactive(sequence_counter):
                # invalid silent Supvisors instances
                self.invalid(status)
                invalidated_identifiers.append(status.identifier)
                # for processes that were running on node, invalidate node in process
                # WARN: Decision is made NOT to remove the node payload from the ProcessStatus and NOT to remove
                #       the ProcessStatus from the Context if no more node payload left.
                #       The aim is to keep a trace in the Web UI about the application processes that have been lost
                #       and their related description.
                failed_processes.update({process for process in status.running_processes()
                                         if process.invalidate_identifier(status.identifier)})
        # trigger the corresponding Supvisors events
        self.publish_process_failures(failed_processes)
        #  return the identifiers of all invalidated Supvisors instances and the processes declared in failure
        return invalidated_identifiers, failed_processes

    def publish_process_failures(self, failed_processes: Set[ProcessStatus]) -> None:
        """ Publish the Supvisors events related with the processes failures.

        :param failed_processes: the processes in failure.
        :return: None.
        """
        # publish process status in failure
        if self.external_publisher:
            for process in failed_processes:
                self.external_publisher.send_process_status(process.serial())
        # update all application sequences and status
        for application_name in {process.application_name for process in failed_processes}:
            application = self.applications[application_name]
            # the application sequences update is useless as long as the application.process map is not impacted
            # (see decision comment above in on_timer_event)
            # application.update_sequences()
            application.update()
            if self.external_publisher:
                self.external_publisher.send_application_status(application.serial())

    def on_instance_failure(self, status: SupvisorsInstanceStatus) -> Set[ProcessStatus]:
        """ Invalid a Supvisors instance that had an XML-RPC failure.

        :param status: the Supvisors instance that sent the event.
        :return: the identifiers of the invalidated Supvisors instances and the processes in failure.
        """
        failed_processes: Set[ProcessStatus] = set()
        # invalid silent Supvisors instances
        self.invalid(status)
        # for processes that were running on node, invalidate node in process
        failed_processes.update({process for process in status.running_processes()
                                 if process.invalidate_identifier(status.identifier)})
        # trigger the corresponding Supvisors events
        self.publish_process_failures(failed_processes)
        return failed_processes

    def check_process(self, status: SupvisorsInstanceStatus,
                      event: Payload, check_source=True) -> Optional[Tuple[ApplicationStatus, ProcessStatus]]:
        """ Check and return the internal data corresponding to the process event.

        :param status: the Supvisors instance from which the event has been received
        :param event: the event payload
        :param check_source: if True, the process should contain information related to the Supvisors instance
        :return: None
        """
        application_name, process_name = event['group'], event['name']
        try:
            application = self.applications[application_name]
            if process_name == '*':
                process = None
            else:
                process = application.processes[process_name]
                assert not check_source or status.identifier in process.info_map
            return application, process
        except (AssertionError, KeyError):
            namespec = make_namespec(application_name, process_name)
            # if the event is received during a Supvisors closing state, it can be ignored
            # otherwise, it means that there's a discrepancy in the internal context, which requires CHECKING
            if self.supvisors.fsm.state in CLOSING_STATES:
                self.logger.warn(f'Context.check_process: ignoring unknown event about process={namespec}'
                                 f' received from Supvisors={status.usage_identifier} while in'
                                 f' {self.supvisors.fsm.state} state')
            else:
                self.logger.error(f'Context.check_process: CHECKING required due to unknown event about'
                                  f' process={namespec} received from Supvisors={status.usage_identifier}')
                self.invalid(status)

    def on_process_removed_event(self, status: SupvisorsInstanceStatus, event: Payload) -> None:
        """ Method called upon reception of a process removed event from the remote Supvisors instance.
        Following an XML-RPC update_numprocs, the size of homogeneous process groups may decrease and lead to removal
        of processes.
        A call to the XML-RPC removeProcessGroup triggers this event too.

        :param status: the Supvisors instance from which the event has been received.
        :param event: the event payload.
        :return: None.
        """
        # accept events only in CHECKED / RUNNING state
        if status.state in [SupvisorsInstanceStates.CHECKED, SupvisorsInstanceStates.RUNNING]:
            self.logger.debug(f'Context.on_remove_process_event: got event {event}'
                              f' from Supvisors={status.usage_identifier}')
            # get internal data
            app_proc = self.check_process(status, event)
            if app_proc:
                application_impacted = False
                # get process targets
                application, event_process = app_proc
                processes = [event_process] if event_process else application.get_instance_processes(status.identifier)
                for process in processes:
                    # WARN: process_failures are not triggered here as the processes have been properly stopped
                    #  as a consequence of the user action
                    # In order to inform users on Supvisors event interface, a fake state DELETED (-1) is sent
                    process_event = event.copy()
                    process_event.update({'name': process.process_name, 'state': -1})
                    if self.external_publisher:
                        self.external_publisher.send_process_event(process_event)
                    # remove process from instance_status
                    status.remove_process(process)
                    # delete the process info entry related to the node
                    if process.remove_identifier(status.identifier):
                        if self.external_publisher:
                            # publish a last process status before it is deleted
                            payload = process.serial()
                            payload.update({'statecode': -1, 'statename': 'DELETED'})
                            self.external_publisher.send_process_status(payload)
                        # there's no more Supvisors instance supporting the process definition
                        # so remove the process from the application
                        application.remove_process(process.process_name)
                        application_impacted = True
                # an update of numprocs cannot leave the application empty (update_numprocs 0 not allowed)
                # however, a remove_group can induce this situation
                if not application.processes:
                    application.state = ApplicationStates.DELETED
                    del self.applications[application.application_name]
                # send an application status when impacted
                if application_impacted and self.external_publisher:
                    self.external_publisher.send_application_status(application.serial())

    def on_process_disability_event(self, status: SupvisorsInstanceStatus, event: Payload) -> None:
        """ Method called upon reception of a process enabled event from the remote Supvisors instance.
        Following an XML-RPC enable/disable on a program, the corresponding process are allowed to be started or not.

        :param status: the Supvisors instance from which the event has been received
        :param event: the event payload
        :return: None
        """
        # accept events only in CHECKED / RUNNING state
        if status.state in [SupvisorsInstanceStates.CHECKED, SupvisorsInstanceStates.RUNNING]:
            self.logger.debug(f'Context.on_process_enabled_event: got event {event}'
                              f' from Supvisors={status.usage_identifier}')
            # get internal data
            app_proc = self.check_process(status, event)
            if app_proc:
                process = app_proc[1]
                # update the process info entry related to the node
                process.update_disability(status.identifier, event['disabled'])
                # at the moment, process disability has no impact on the application and process status
                # so only the process event publication makes sense
                if self.external_publisher:
                    self.external_publisher.send_process_event(event)

    def on_process_state_event(self, status: SupvisorsInstanceStatus, event: Payload) -> Optional[ProcessStatus]:
        """ Method called upon reception of a process event from the remote Supvisors instance.
        Supvisors checks that the handling of the event is valid in case of auto fencing.
        The method updates the ProcessStatus corresponding to the event, and thus the wrapping ApplicationStatus.
        Finally, the updated ProcessStatus and ApplicationStatus are published.

        :param status: the Supvisors instance from which the event has been received.
        :param event: the event payload.
        :return: None.
        """
        # accept events only in CHECKED / RUNNING state
        if status.state in [SupvisorsInstanceStates.CHECKED, SupvisorsInstanceStates.RUNNING]:
            self.logger.debug(f'Context.on_process_event: got event {event} from Supvisors={status.usage_identifier}')
            # WARN: the Master may send a process event corresponding to a process that is not configured in it
            forced_event = 'forced' in event
            app_proc = self.check_process(status, event, not forced_event)
            if not app_proc:
                self.logger.trace('Context.on_process_event: could not find any process corresponding'
                                  f' to event={event}')
            else:
                application, process = app_proc
                updated = True
                # refresh process info depending on the nature of the process event
                if forced_event:
                    updated = process.force_state(event)
                    self.logger.trace(f'Context.on_process_event: {process.namespec} forced event'
                                      f' considered={updated}')
                    if updated:
                        # remove the 'forced' status before publication
                        # NOTE: use a copy so that the caller is not impacted by the payload change
                        event = event.copy()
                        event['state'] = process.displayed_state
                        del event['forced']
                else:
                    # update the ProcessStatus based on new information received from a local Supvisors instance
                    process.update_info(status.identifier, event)
                    try:
                        # update command line in Supervisor
                        self.supvisors.supervisor_data.update_extra_args(process.namespec, event['extra_args'])
                    except KeyError:
                        # process not found in Supervisor internal structure
                        self.logger.debug(f'Context.on_process_event: cannot apply extra args to {process.namespec}'
                                          ' unknown to local Supervisor')
                # forced event may be dismissed
                if updated:
                    # refresh internal status
                    status.update_process(process)
                    application.update()
                    # publish process event, status and application status
                    if self.external_publisher:
                        self.external_publisher.send_process_event(event)
                        self.external_publisher.send_process_status(process.serial())
                        self.external_publisher.send_application_status(application.serial())
                    return process



# ======================================================================
# Copyright 2016 Julien LE CLEACH
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

import random
import sys
from socket import getfqdn
from unittest.mock import call

import pytest

from supvisors.application import *
from .base import database_copy, process_info_by_name, any_stopped_process_info, any_process_info_by_state
from .conftest import create_application, create_process


# ApplicationRules part
@pytest.fixture
def rules(supvisors):
    """ Return the instance to test. """
    return ApplicationRules(supvisors)


def test_rules_create(rules):
    """ Test the values set at construction. """
    # check application default rules
    assert not rules.managed
    assert rules.distribution == DistributionRules.ALL_INSTANCES
    assert rules.identifiers == ['*']
    assert rules.hash_identifiers == []
    assert rules.start_sequence == 0
    assert rules.stop_sequence == -1
    assert rules.starting_strategy == StartingStrategies.CONFIG
    assert rules.starting_failure_strategy == StartingFailureStrategies.ABORT
    assert rules.running_failure_strategy == RunningFailureStrategies.CONTINUE
    assert rules.status_tree is None


def test_rules_check_stop_sequence(rules):
    """ Test the assignment of stop sequence to start sequence if default still set. """
    # test when default still used
    assert rules.start_sequence == 0
    assert rules.stop_sequence == -1
    rules.check_stop_sequence('crash')
    assert rules.start_sequence == 0
    assert rules.stop_sequence == 0
    # test when value has been set
    rules.start_sequence = 12
    rules.stop_sequence = 50
    rules.check_stop_sequence('crash')
    assert rules.start_sequence == 12
    assert rules.stop_sequence == 50


def test_rules_check_hash_identifiers(rules):
    """ Test the resolution of identifiers when hash_identifiers is set. """
    # set initial attributes
    rules.hash_identifiers = ['*']
    rules.identifiers = []
    rules.start_sequence = 1
    # 1. test with application without ending index
    rules.check_hash_identifiers('crash')
    # identifiers is unchanged and start_sequence is invalidated
    assert rules.hash_identifiers == ['*']
    assert rules.identifiers == []
    assert rules.start_sequence == 0
    # 2. test with application with 0-ending index
    rules.start_sequence = 1
    rules.check_hash_identifiers('sample_test_0')
    # identifiers is unchanged and start_sequence is invalidated
    assert rules.hash_identifiers == ['*']
    assert rules.identifiers == []
    assert rules.start_sequence == 0
    # 3. update rules to test '#' with all instances available
    # address '10.0.0.1' has an index of 1-1 in supvisors_mapper
    rules.start_sequence = 1
    rules.check_hash_identifiers('sample_test_1')
    assert rules.identifiers == ['10.0.0.1:25000']
    assert rules.start_sequence == 1
    # 4. update rules to test '#' with a subset of instances available
    rules.hash_identifiers = ['10.0.0.0:25000', '10.0.0.3:25000', '10.0.0.5:25000']
    rules.identifiers = []
    # here, at index 2-1 of this list, '10.0.0.5' can be found
    rules.check_hash_identifiers('sample_test_2')
    assert rules.identifiers == ['10.0.0.3:25000']
    assert rules.start_sequence == 1
    # 5. test the case where procnumber is greater than the subset list of instances available
    rules.hash_identifiers = ['10.0.0.1:25000']
    rules.identifiers = []
    rules.check_hash_identifiers('sample_test_2')
    assert rules.identifiers == ['10.0.0.1:25000']
    assert rules.start_sequence == 1


def test_rules_check_dependencies(mocker, rules):
    """ Test the dependencies in process rules. """
    mocked_stop = mocker.patch('supvisors.application.ApplicationRules.check_stop_sequence')
    mocked_hash = mocker.patch('supvisors.application.ApplicationRules.check_hash_identifiers')
    # test with no hash
    rules.hash_identifiers = []
    rules.check_dependencies('dummy')
    assert mocked_stop.call_args_list == [call('dummy')]
    assert not mocked_hash.called
    mocker.resetall()
    # test with hash
    rules.hash_identifiers = ['*']
    rules.check_dependencies('dummy')
    assert mocked_stop.call_args_list == [call('dummy')]
    assert mocked_hash.call_args_list == [call('dummy')]


def test_rules_str(rules):
    """ Test the string output. """
    assert str(rules) == ("managed=False distribution=ALL_INSTANCES identifiers=['*'] start_sequence=0 stop_sequence=-1"
                          " starting_strategy=CONFIG starting_failure_strategy=ABORT running_failure_strategy=CONTINUE"
                          " status_formula=None")


def test_rules_serial(rules):
    """ Test the serialization of the ApplicationRules object. """
    # default is not managed so result is short
    assert rules.serial() == {'managed': False}
    # check managed and distributed
    rules.managed = True
    assert rules.serial() == {'managed': True, 'distribution': 'ALL_INSTANCES', 'identifiers': ['*'],
                              'start_sequence': 0, 'stop_sequence': -1,
                              'starting_strategy': 'CONFIG', 'starting_failure_strategy': 'ABORT',
                              'running_failure_strategy': 'CONTINUE',
                              'status_formula': ''}
    # finally check managed and not distributed
    rules.distribution = DistributionRules.SINGLE_INSTANCE
    rules.status_formula = 'dumb and dumber'
    assert rules.serial() == {'managed': True, 'distribution': 'SINGLE_INSTANCE', 'identifiers': ['*'],
                              'start_sequence': 0, 'stop_sequence': -1,
                              'starting_strategy': 'CONFIG', 'starting_failure_strategy': 'ABORT',
                              'running_failure_strategy': 'CONTINUE',
                              'status_formula': 'dumb and dumber'}


# Homogeneous group part
def test_homogeneous_group_create(supvisors):
    """ Test the values set at construction. """
    group = HomogeneousGroup('yeux', supvisors)
    assert group.supvisors is supvisors
    assert group.program_name == 'yeux'
    assert group.at_identifiers is None
    assert group.hash_identifiers is None
    assert group.processes == []
    assert group.logger is supvisors.logger


def test_homogeneous_group_add_remove(supvisors):
    """ Test the HomogeneousGroup methods add_process and remove_process. """
    group = HomogeneousGroup('yeux', supvisors)
    # get 2 processes for test
    info = process_info_by_name('yeux_00')
    process_1 = create_process(info, supvisors)
    process_1.add_info('10.0.0.1', info)
    info = process_info_by_name('yeux_01')
    process_2 = create_process(info, supvisors)
    process_2.add_info('10.0.0.1', info)
    # 1. add a process with default rules to the application
    group.add_process(process_1)
    # check that process is stored
    assert process_1 in group.processes
    assert group.at_identifiers is None
    assert group.hash_identifiers is None
    # 2. test removal
    group.remove_process(process_1)
    # check that process is not stored anymore
    assert group.processes == []
    assert group.at_identifiers is None
    assert group.hash_identifiers is None
    # 3. add a process with hash rules to the application
    process_1.rules.hash_identifiers = ['10.0.0.1']
    group.add_process(process_1)
    # check that process is stored
    assert group.processes == [process_1]
    assert group.at_identifiers is None
    assert group.hash_identifiers == ['10.0.0.1']
    # 4. add a process with same program name and at rules to the application
    process_2.rules.hash_identifiers = ['10.0.0.2', '10.0.0.3']
    group.add_process(process_2)
    # check that process is stored
    assert group.processes == [process_1, process_2]
    assert group.at_identifiers is None
    assert group.hash_identifiers == ['10.0.0.2', '10.0.0.3']
    # 5. test removal
    group.remove_process(process_1)
    # check that process is not stored anymore
    assert group.processes == [process_2]
    assert group.at_identifiers is None
    assert group.hash_identifiers == ['10.0.0.2', '10.0.0.3']
    # 6. test removal
    group.remove_process(process_2)
    # check that process is not stored anymore
    assert group.processes == []
    assert group.at_identifiers is None
    assert group.hash_identifiers == ['10.0.0.2', '10.0.0.3']
    # 7. quickly test with at_identifiers (at prevail on hash)
    process_1.rules.at_identifiers = ['10.0.0.1']
    process_1.rules.hash_identifiers = []
    group.add_process(process_1)
    # check stored information
    assert group.processes == [process_1]
    assert group.at_identifiers == ['10.0.0.1']
    assert group.hash_identifiers is None
    # 8. second process with at identifier
    process_2.rules.at_identifiers = ['10.0.0.0']
    process_2.rules.hash_identifiers = []
    group.add_process(process_2)
    # check stored information
    assert group.processes == [process_1, process_2]
    assert group.at_identifiers == ['10.0.0.0']
    assert group.hash_identifiers is None


@pytest.fixture
def homogeneous_group(supvisors):
    """ Return the HomogeneousGroup instance to test. """
    group = HomogeneousGroup('yeux', supvisors)
    # add 2 processes of the same program
    info = process_info_by_name('yeux_00')
    process_1 = create_process(info, supvisors)
    process_1.add_info('10.0.0.1:25000', info)
    group.add_process(process_1)
    info = process_info_by_name('yeux_01')
    process_2 = create_process(info, supvisors)
    process_2.add_info('10.0.0.1:25000', info)
    group.add_process(process_2)
    return group


def test_homogeneous_group_resolve_at_wildcard(homogeneous_group):
    """ Test the HomogeneousGroup.assign_at_identifiers method with wildcard rules and no discovery mode. """
    # force at-* rules in processes and group
    homogeneous_group.at_identifiers = [WILDCARD]
    process_1 = homogeneous_group.processes[0]
    process_1.rules.at_identifiers = [WILDCARD]
    process_1.rules.identifiers = []
    process_2 = homogeneous_group.processes[1]
    process_2.rules.at_identifiers = [WILDCARD]
    process_2.rules.identifiers = []
    # 1. check assignment
    # the number of instance exceeds the number of processes
    homogeneous_group.resolve_rules()
    assert process_1.rules.at_identifiers == []
    assert process_1.rules.identifiers == ['10.0.0.1:25000']
    assert process_2.rules.at_identifiers == []
    assert process_2.rules.identifiers == ['10.0.0.2:25000']
    # 2. check no change with same call
    homogeneous_group.resolve_rules()
    assert process_1.rules.at_identifiers == []
    assert process_1.rules.identifiers == ['10.0.0.1:25000']
    assert process_2.rules.at_identifiers == []
    assert process_2.rules.identifiers == ['10.0.0.2:25000']
    # 3. in standard mode, the list of Supvisors instances cannot change
    #    but the number of processes in a homogeneous group can (update_numprocs XML-RPC)
    info = process_2.info_map['10.0.0.1:25000']
    info.update({'name': 'yeux_02', 'process_index': 2})
    process_3 = create_process(info, homogeneous_group.supvisors)
    process_3.add_info('10.0.0.1:25000', info)
    process_3.rules.at_identifiers = [WILDCARD]
    process_3.rules.identifiers = []
    homogeneous_group.add_process(process_3)
    # call new resolution
    homogeneous_group.resolve_rules()
    assert process_1.rules.at_identifiers == []
    assert process_1.rules.identifiers == ['10.0.0.1:25000']
    assert process_2.rules.at_identifiers == []
    assert process_2.rules.identifiers == ['10.0.0.2:25000']
    assert process_3.rules.at_identifiers == []
    assert process_3.rules.identifiers == ['10.0.0.3:25000']


def test_homogeneous_group_resolve_at_list(homogeneous_group):
    """ Test the HomogeneousGroup.assign_at_identifiers method. """
    # patch list of instances so that it is smaller than the list of processes
    mapper = homogeneous_group.supvisors.mapper
    ref_instances = mapper.instances.copy()
    mapper._instances = {'10.0.0.1:25000': ref_instances['10.0.0.1:25000']}
    # force at-* rules in processes and group
    homogeneous_group.at_identifiers = ['10.0.0.2:25000', '10.0.0.1:25000']
    process_1 = homogeneous_group.processes[0]
    process_1.rules.at_identifiers = ['10.0.0.2:25000', '10.0.0.1:25000']
    process_1.rules.identifiers = []
    process_2 = homogeneous_group.processes[1]
    process_2.rules.at_identifiers = ['10.0.0.2:25000', '10.0.0.1:25000']
    process_2.rules.identifiers = []
    # 1. check assignment
    # the number of instance exceeds the number of processes
    homogeneous_group.resolve_rules()
    assert process_1.rules.at_identifiers == []
    assert process_1.rules.identifiers == ['10.0.0.1:25000']
    assert process_2.rules.at_identifiers == ['10.0.0.2:25000', '10.0.0.1:25000']
    assert process_2.rules.identifiers == []
    # 2. check no change with same call
    homogeneous_group.resolve_rules()
    assert process_1.rules.at_identifiers == []
    assert process_1.rules.identifiers == ['10.0.0.1:25000']
    assert process_2.rules.at_identifiers == ['10.0.0.2:25000', '10.0.0.1:25000']
    assert process_2.rules.identifiers == []
    # 3. in discovery mode, the list of Supvisors instances may increase
    mapper._instances = {'10.0.0.1:25000': ref_instances['10.0.0.1:25000'],
                         '10.0.0.3:25000': ref_instances['10.0.0.3:25000'],
                         '10.0.0.2:25000': ref_instances['10.0.0.2:25000']}
    # call new resolution
    homogeneous_group.resolve_rules()
    assert process_1.rules.at_identifiers == []
    assert process_1.rules.identifiers == ['10.0.0.1:25000']
    assert process_2.rules.at_identifiers == []
    assert process_2.rules.identifiers == ['10.0.0.2:25000']


def test_homogeneous_group_resolve_hash_wildcard(homogeneous_group):
    """ Test the HomogeneousGroup.assign_at_identifiers method with wildcard rules and no discovery mode. """
    # patch list of instances so that it is smaller than the list of processes
    mapper = homogeneous_group.supvisors.mapper
    ref_instances = mapper.instances.copy()
    mapper._instances = {'10.0.0.1:25000': ref_instances['10.0.0.1:25000'],
                         '10.0.0.2:25000': ref_instances['10.0.0.2:25000']}
    # force hash-* rules in processes and group
    homogeneous_group.hash_identifiers = [WILDCARD]
    process_1 = homogeneous_group.processes[0]
    process_1.rules.hash_identifiers = [WILDCARD]
    process_1.rules.identifiers = []
    process_2 = homogeneous_group.processes[1]
    process_2.rules.hash_identifiers = [WILDCARD]
    process_2.rules.identifiers = []
    # 1. check assignment
    # the number of instance exceeds the number of processes
    homogeneous_group.resolve_rules()
    assert process_1.rules.hash_identifiers == []
    assert process_1.rules.identifiers == ['10.0.0.1:25000']
    assert process_2.rules.hash_identifiers == []
    assert process_2.rules.identifiers == ['10.0.0.2:25000']
    # 2. check no change with same call
    homogeneous_group.resolve_rules()
    assert process_1.rules.hash_identifiers == []
    assert process_1.rules.identifiers == ['10.0.0.1:25000']
    assert process_2.rules.hash_identifiers == []
    assert process_2.rules.identifiers == ['10.0.0.2:25000']
    # 3. in standard mode, the list of Supvisors instances cannot change
    #    but the number of processes in a homogeneous group can (update_numprocs XML-RPC)
    info = process_2.info_map['10.0.0.1:25000']
    info.update({'name': 'yeux_02', 'process_index': 2})
    process_3 = create_process(info, homogeneous_group.supvisors)
    process_3.add_info('10.0.0.1:25000', info)
    process_3.rules.hash_identifiers = [WILDCARD]
    process_3.rules.identifiers = []
    homogeneous_group.add_process(process_3)
    # call new resolution
    homogeneous_group.resolve_rules()
    assert process_1.rules.hash_identifiers == []
    assert process_1.rules.identifiers == ['10.0.0.1:25000']
    assert process_2.rules.hash_identifiers == []
    assert process_2.rules.identifiers == ['10.0.0.2:25000']
    assert process_3.rules.hash_identifiers == []
    assert process_3.rules.identifiers == ['10.0.0.1:25000']


def test_homogeneous_group_resolve_hash_list(homogeneous_group):
    """ Test the HomogeneousGroup.assign_at_identifiers method. """
    # patch list of instances so that it is smaller than the list of processes
    mapper = homogeneous_group.supvisors.mapper
    ref_instances = mapper.instances.copy()
    mapper._instances = {'10.0.0.1:25000': ref_instances['10.0.0.1:25000'],
                         '10.0.0.2:25000': ref_instances['10.0.0.2:25000']}
    # force at-* rules in processes and group
    homogeneous_group.hash_identifiers = ['10.0.0.3:25000', '10.0.0.2:25000', '10.0.0.1:25000']
    process_1 = homogeneous_group.processes[0]
    process_1.rules.hash_identifiers = ['10.0.0.3:25000', '10.0.0.2:25000', '10.0.0.1:25000']
    process_1.rules.identifiers = []
    process_2 = homogeneous_group.processes[1]
    process_2.rules.hash_identifiers = ['10.0.0.3:25000', '10.0.0.2:25000', '10.0.0.1:25000']
    process_2.rules.identifiers = []
    # 1. check assignment
    # the number of instance exceeds the number of processes
    homogeneous_group.resolve_rules()
    assert process_1.rules.hash_identifiers == []
    assert process_1.rules.identifiers == ['10.0.0.2:25000']
    assert process_2.rules.hash_identifiers == []
    assert process_2.rules.identifiers == ['10.0.0.1:25000']
    # 2. check no change with same call
    homogeneous_group.resolve_rules()
    assert process_1.rules.hash_identifiers == []
    assert process_1.rules.identifiers == ['10.0.0.2:25000']
    assert process_2.rules.hash_identifiers == []
    assert process_2.rules.identifiers == ['10.0.0.1:25000']
    # 3. in standard mode, the list of Supvisors instances cannot change
    #    but the number of processes in a homogeneous group can (update_numprocs XML-RPC)
    info = process_2.info_map['10.0.0.1:25000']
    info.update({'name': 'yeux_02', 'process_index': 2})
    process_3 = create_process(info, homogeneous_group.supvisors)
    process_3.add_info('10.0.0.1:25000', info)
    process_3.rules.hash_identifiers = ['10.0.0.3:25000', '10.0.0.2:25000', '10.0.0.1:25000']
    process_3.rules.identifiers = []
    homogeneous_group.add_process(process_3)
    # call new resolution
    homogeneous_group.resolve_rules()
    assert process_1.rules.hash_identifiers == []
    assert process_1.rules.identifiers == ['10.0.0.2:25000']
    assert process_2.rules.hash_identifiers == []
    assert process_2.rules.identifiers == ['10.0.0.1:25000']
    assert process_3.rules.hash_identifiers == []
    assert process_3.rules.identifiers == ['10.0.0.2:25000']


# ApplicationStatus part
def test_application_create(supvisors):
    """ Test the values set at construction. """
    application = create_application('ApplicationTest', supvisors)
    # check application default attributes
    assert application.supvisors is supvisors
    assert application.logger is supvisors.logger
    assert application.application_name == 'ApplicationTest'
    assert application.state == ApplicationStates.STOPPED
    assert not application.major_failure
    assert not application.minor_failure
    assert not application.processes
    assert not application.start_sequence
    assert not application.stop_sequence
    # check application default rules
    assert not application.rules.managed
    assert application.rules.distribution == DistributionRules.ALL_INSTANCES
    assert application.rules.start_sequence == 0
    assert application.rules.stop_sequence == -1
    assert application.rules.starting_failure_strategy == StartingFailureStrategies.ABORT
    assert application.rules.running_failure_strategy == RunningFailureStrategies.CONTINUE


def test_application_running(supvisors):
    """ Test the running method. """
    application = create_application('ApplicationTest', supvisors)
    assert not application.running()
    # loop on all states
    for state in ApplicationStates:
        application._state = state
        assert application.running() == (state in [ApplicationStates.STARTING, ApplicationStates.RUNNING])


def test_application_stopped(supvisors):
    """ Test the ApplicationStatus.stopped method. """
    application = create_application('ApplicationTest', supvisors)
    assert application.stopped()
    # loop on all states
    for state in ApplicationStates:
        application._state = state
        assert application.stopped() == (state == ApplicationStates.STOPPED)


def test_application_never_started(supvisors):
    """ Test the ApplicationStatus.never_started method. """
    application = create_application('ApplicationTest', supvisors)
    assert application.never_started()
    # add a stopped process that has already been started
    info = any_stopped_process_info()
    process = create_process(info, supvisors)
    process.add_info('10.0.0.1', info)
    application.add_process(process)
    application.update_state()
    assert not application.never_started()


def test_application_has_running_processes(supvisors):
    """ Test the ApplicationStatus.has_running_processes method used to know if at least one process is RUNNING. """
    application = create_application('ApplicationTest', supvisors)
    assert not application.has_running_processes()
    # add a stopped process
    info = any_stopped_process_info()
    process = create_process(info, supvisors)
    process.add_info('10.0.0.1', info)
    application.add_process(process)
    application.update()
    assert application.stopped()
    assert not application.has_running_processes()
    # add a running process
    info = any_process_info_by_state(ProcessStates.RUNNING)
    process = create_process(info, supvisors)
    process.add_info('10.0.0.1', info)
    application.add_process(process)
    application.update_state()
    assert application.stopped()
    assert application.has_running_processes()


def test_application_get_operational_status(supvisors):
    """ Test the ApplicationStatus.get_operational_status method used to get a descriptive operational status. """
    # create address status instance
    application = create_application('ApplicationTest', supvisors)
    # test with non RUNNING application
    for state in [ApplicationStates.STOPPED, ApplicationStates.STARTING, ApplicationStates.STOPPING]:
        for minor_failure in [True, False]:
            for major_failure in [True, False]:
                application._state = state
                application.major_failure = major_failure
                application.minor_failure = minor_failure
                application.start_failure = False
                assert application.get_operational_status() == ''
    # test with RUNNING application
    application._state = ApplicationStates.RUNNING
    # no failure
    application.major_failure = False
    application.minor_failure = False
    application.start_failure = False
    assert application.get_operational_status() == 'Operational'
    # minor failure, no major failure
    application.major_failure = False
    application.minor_failure = True
    application.start_failure = False
    assert application.get_operational_status() == 'Degraded'
    # major failure set
    application.major_failure = True
    application.start_failure = False
    for minor_failure in [True, False]:
        application.minor_failure = minor_failure
        assert application.get_operational_status() == 'Not Operational'


def test_application_serial(supvisors):
    """ Test the serial method used to get a serializable form of Application. """
    import pickle
    # create address status instance
    application = create_application('ApplicationTest', supvisors)
    application._state = ApplicationStates.RUNNING
    application.major_failure = False
    application.minor_failure = True
    application.start_failure = False
    # test to_json method
    serialized = application.serial()
    assert serialized == {'application_name': 'ApplicationTest', 'managed': False,
                          'statecode': 2, 'statename': 'RUNNING', 'major_failure': False, 'minor_failure': True}
    # test that returned structure is serializable using pickle
    dumped = pickle.dumps(serialized)
    loaded = pickle.loads(dumped)
    assert serialized == loaded


def test_application_add_remove_process(supvisors):
    """ Test the add_process and remove_process methods. """
    application = create_application('ApplicationTest', supvisors)
    # get 3 processes for test
    processes = []
    for name in ['xclock', 'yeux_00', 'yeux_01']:
        info = process_info_by_name(name)
        process = create_process(info, supvisors)
        process.add_info('10.0.0.1', info)
        processes.append(process)
    process_1 = processes[0]
    process_2 = processes[1]
    process_2.rules.at_identifiers = ['10.0.0.1']
    process_3 = processes[2]
    process_3.program_name = process_2.program_name
    process_3.rules.at_identifiers = ['10.0.0.2', '10.0.0.3']
    # 1. add a process with default rules to the application
    application.add_process(process_1)
    # check that process is stored
    assert application.processes == {process_1.process_name: process_1}
    assert list(application.process_groups.keys()) == [process_1.program_name]
    group = application.process_groups[process_1.program_name]
    assert process_1 in group.processes
    assert group.at_identifiers is None
    assert group.hash_identifiers is None
    # 2. add a process with hash rules to the application
    application.add_process(process_2)
    # check that process is stored
    assert application.processes == {process_1.process_name: process_1,
                                     process_2.process_name: process_2}
    assert sorted(application.process_groups.keys()) == sorted([process_1.program_name, process_2.program_name])
    group = application.process_groups[process_2.program_name]
    assert group.processes == [process_2]
    assert group.at_identifiers == ['10.0.0.1']
    assert group.hash_identifiers is None
    # 3. add a process with same program name and at rules to the application
    application.add_process(process_3)
    # check that process is stored
    assert application.processes == {process_1.process_name: process_1,
                                     process_2.process_name: process_2,
                                     process_3.process_name: process_3}
    assert sorted(application.process_groups.keys()) == sorted([process_1.program_name, process_2.program_name])
    group = application.process_groups[process_3.program_name]
    assert group.processes == [process_2, process_3]
    assert group.at_identifiers == ['10.0.0.2', '10.0.0.3']
    assert group.hash_identifiers is None
    # 4. test first removal
    application.remove_process(process_2.process_name)
    # check that process is not stored anymore
    assert application.processes == {process_1.process_name: process_1,
                                     process_3.process_name: process_3}
    assert sorted(application.process_groups.keys()) == sorted([process_1.program_name, process_2.program_name])
    group = application.process_groups[process_1.program_name]
    assert group.processes == [process_1]
    group = application.process_groups[process_3.program_name]
    assert group.processes == [process_3]
    # 5. test second removal
    application.remove_process(process_1.process_name)
    # check that process is not stored anymore
    assert application.processes == {process_3.process_name: process_3}
    assert list(application.process_groups.keys()) == [process_3.program_name]
    group = application.process_groups[process_3.program_name]
    assert group.processes == [process_3]
    # 5. test third removal
    application.remove_process(process_3.process_name)
    # check that process is not stored anymore
    assert application.processes == {}
    assert application.process_groups == {}
    # 6. quickly test with hash_identifiers
    process_2.rules.at_identifiers = []
    process_2.rules.hash_identifiers = ['10.0.0.1']
    process_3.rules.at_identifiers = []
    process_3.rules.hash_identifiers = ['10.0.0.2', '10.0.0.3']
    application.add_process(process_3)
    application.add_process(process_2)
    # check stored information
    assert application.processes == {process_2.process_name: process_2,
                                     process_3.process_name: process_3}
    assert list(application.process_groups.keys()) == [process_2.program_name]
    group = application.process_groups[process_3.program_name]
    assert group.processes == [process_3, process_2]
    assert group.at_identifiers is None
    assert group.hash_identifiers == ['10.0.0.1']
    # 7. finish with last "should not happen" case
    process_1.program_name = process_2.program_name
    process_1.rules.at_identifiers = ['10.0.0.0']
    application.add_process(process_1)
    # check stored information
    assert application.processes == {process_1.process_name: process_1,
                                     process_2.process_name: process_2,
                                     process_3.process_name: process_3}
    assert list(application.process_groups.keys()) == [process_2.program_name]
    group = application.process_groups[process_1.program_name]
    assert group.processes == [process_3, process_2, process_1]
    assert group.at_identifiers == ['10.0.0.0']
    assert group.hash_identifiers is None


def test_application_possible_identifiers(supvisors):
    """ Test the ApplicationStatus.possible_identifiers method. """
    application = create_application('ApplicationTest', supvisors)
    # add a process to the application
    info = any_process_info_by_state(ProcessStates.STARTING)
    process1 = create_process(info, supvisors)
    for node_name in ['10.0.0.2:25000', '10.0.0.3:25000', '10.0.0.4:25000']:
        process1.add_info(node_name, info.copy())
    application.add_process(process1)
    # add another process to the application
    info = any_stopped_process_info()
    process2 = create_process(info, supvisors)
    for node_name in ['10.0.0.1:25000', '10.0.0.4:25000']:
        process2.add_info(node_name, info.copy())
    application.add_process(process2)
    # default identifiers is '*' in process rules
    assert application.possible_identifiers() == ['10.0.0.4:25000']
    # set a subset of identifiers in process rules so that there's no intersection with received status
    application.rules.identifiers = ['10.0.0.1:25000', '10.0.0.2:25000']
    assert application.possible_identifiers() == []
    # increase received status
    process1.add_info('10.0.0.1:25000', info.copy())
    assert application.possible_identifiers() == ['10.0.0.1:25000']
    # disable program on '10.0.0.1'
    process2.update_disability('10.0.0.1:25000', True)
    assert application.possible_identifiers() == []
    # reset rules
    application.rules.identifiers = ['*']
    assert application.possible_identifiers() == ['10.0.0.4:25000']
    # test with full status and all instances in rules + re-enable on '10.0.0.1'
    process2.update_disability('10.0.0.1:25000', False)
    for node_name in supvisors.mapper.instances:
        process1.add_info(node_name, info.copy())
        process2.add_info(node_name, info.copy())
    assert sorted(application.possible_identifiers()) == sorted(supvisors.mapper.instances.keys())
    # restrict again instances in rules
    application.rules.identifiers = ['10.0.0.2:25000', '10.0.0.5:25000']
    assert application.possible_identifiers() == ['10.0.0.2:25000', '10.0.0.5:25000']


def test_application_possible_node_identifiers(supvisors):
    """ Test the ApplicationStatus.possible_node_identifiers method.
    Same test logic as above but update the node mapping before. """
    # update the node mapping
    fqdn = getfqdn()
    supvisors.mapper._nodes = {'10.0.0.1': ['10.0.0.1:25000', '10.0.0.3:25000', '10.0.0.5:25000'],
                               '10.0.0.2': ['10.0.0.2:25000', '10.0.0.4:25000'],
                               fqdn: [f'{fqdn}:25000', f'{fqdn}:15000']}
    # create the test application
    application = create_application('ApplicationTest', supvisors)
    # add a process to the application
    info = any_process_info_by_state(ProcessStates.STARTING)
    process1 = create_process(info, supvisors)
    for identifier in ['10.0.0.2:25000', '10.0.0.3:25000', '10.0.0.4:25000']:
        process1.add_info(identifier, info.copy())
    application.add_process(process1)
    # add another process to the application
    info = any_stopped_process_info()
    process2 = create_process(info, supvisors)
    for identifier in ['10.0.0.1:25000', '10.0.0.4:25000']:
        process2.add_info(identifier, info.copy())
    application.add_process(process2)
    # default identifiers is '*' in process rules
    assert application.possible_node_identifiers() == ['10.0.0.1:25000', '10.0.0.2:25000',
                                                       '10.0.0.3:25000', '10.0.0.4:25000']
    # set a subset of identifiers in process rules so that there's no intersection with received status
    application.rules.identifiers = ['10.0.0.1:25000', '10.0.0.2:25000']
    assert application.possible_node_identifiers() == []
    # increase received status
    process1.add_info('10.0.0.1:25000', info.copy())
    assert application.possible_node_identifiers() == ['10.0.0.1:25000']
    # disable program on 10.0.0.1
    process2.update_disability('10.0.0.1:25000', True)
    assert application.possible_node_identifiers() == []
    # reset rules
    application.rules.identifiers = ['*']
    assert application.possible_node_identifiers() == ['10.0.0.2:25000', '10.0.0.4:25000']
    # test with full status and all instances in rules + re-enable on '10.0.0.1'
    process2.update_disability('10.0.0.1:25000', False)
    for identifier in supvisors.mapper.instances:
        process1.add_info(identifier, info.copy())
        process2.add_info(identifier, info.copy())
    assert sorted(application.possible_node_identifiers()) == sorted(supvisors.mapper.instances.keys())
    # restrict again instances in rules
    application.rules.identifiers = ['10.0.0.2:25000', '10.0.0.5:25000']
    assert application.possible_node_identifiers() == ['10.0.0.2:25000', '10.0.0.5:25000']


def test_application_get_instance_processes(supvisors):
    """ Test the ApplicationStatus.get_instance_processes method. """
    application = create_application('ApplicationTest', supvisors)
    # add a process to the application
    info = any_process_info_by_state(ProcessStates.STARTING)
    process1 = create_process(info, supvisors)
    for node_name in ['10.0.0.2', '10.0.0.3', '10.0.0.4']:
        process1.add_info(node_name, info.copy())
    application.add_process(process1)
    # add another process to the application
    info = any_stopped_process_info()
    process2 = create_process(info, supvisors)
    for node_name in ['10.0.0.1', '10.0.0.4']:
        process2.add_info(node_name, info.copy())
    application.add_process(process2)
    # test call
    assert application.get_instance_processes('10.0.0.0') == []
    assert application.get_instance_processes('10.0.0.1') == [process2]
    assert application.get_instance_processes('10.0.0.2') == [process1]
    assert application.get_instance_processes('10.0.0.3') == [process1]
    assert application.get_instance_processes('10.0.0.4') == [process1, process2]


@pytest.fixture
def filled_application(supvisors):
    """ Create an ApplicationStatus and add all processes of the database. """
    application = create_application('ApplicationTest', supvisors)
    for info in database_copy():
        process = create_process(info, supvisors)
        process.add_info('10.0.0.1', info)
        # set random sequence to process
        process.rules.start_sequence = random.randint(0, 2)
        process.rules.stop_sequence = random.randint(0, 2)
        application.add_process(process)
    return application


def test_application_update_sequences(filled_application):
    """ Test the sequencing of the update_sequences method. """
    # call the sequencer
    filled_application.update_sequences()
    # check the sequencing of the starting
    sequences = sorted({process.rules.start_sequence for process in filled_application.processes.values()})
    # by default, applications are unmanaged so start sequence is empty
    assert not filled_application.start_sequence
    assert filled_application.stop_sequence
    # stop sequence contents is tested afterwards
    # force application to managed and call sequencer again
    filled_application.rules.managed = True
    filled_application.update_sequences()
    # as key is an integer, the sequence dictionary should be sorted but doesn't work in Travis-CI
    assert filled_application.start_sequence
    assert filled_application.stop_sequence
    for sequence, processes in sorted(filled_application.start_sequence.items()):
        assert sequence == sequences.pop(0)
        assert sorted(processes, key=lambda x: x.process_name) == \
               sorted([proc for proc in filled_application.processes.values()
                       if sequence == proc.rules.start_sequence], key=lambda x: x.process_name)
    # check the sequencing of the stopping
    sequences = sorted({process.rules.stop_sequence for process in filled_application.processes.values()})
    # as key is an integer, the sequence dictionary should be sorted but doesn't work in Travis-CI
    for sequence, processes in sorted(filled_application.stop_sequence.items()):
        assert sequence == sequences.pop(0)
        assert sorted(processes, key=lambda x: x.process_name) == \
               sorted([proc for proc in filled_application.processes.values()
                       if sequence == proc.rules.stop_sequence], key=lambda x: x.process_name)


def test_application_get_start_sequence_expected_load(filled_application):
    """ Test the ApplicationStatus.get_start_sequence_expected_load method. """
    # as sequences are empty, total is 0
    assert filled_application.get_start_sequence_expected_load() == 0
    # set application to managed, update sequences and status
    filled_application.rules.managed = True
    filled_application.update_sequences()
    # process rules still have a expected_load set to 0
    assert filled_application.get_start_sequence_expected_load() == 0
    print(ApplicationStatus.printable_sequence(filled_application.start_sequence))
    # update all process loads to 10
    for proc_list in filled_application.start_sequence.values():
        for process in proc_list:
            process.rules.expected_load = 10
    seq_0_size = len(filled_application.start_sequence.get(0, []))
    seq_1_2_size = len(filled_application.processes) - seq_0_size
    assert filled_application.get_start_sequence_expected_load() == 10 * seq_1_2_size


def test_update(mocker, filled_application):
    """ Test the ApplicationStatus.update method. """
    mocked_state = mocker.patch.object(filled_application, 'update_state', return_value=ApplicationStates.RUNNING)
    mocked_formula = mocker.patch.object(filled_application, 'update_status_formula')
    mocked_required = mocker.patch.object(filled_application, 'update_status_required')
    # initial state
    assert filled_application.state == ApplicationStates.STOPPED
    assert not filled_application.major_failure
    assert not filled_application.minor_failure
    # update state and status
    # test with no formula set in application rules
    assert filled_application.rules.status_formula is None
    filled_application.update()
    assert filled_application.state == ApplicationStates.RUNNING
    assert not filled_application.major_failure
    assert not filled_application.minor_failure
    assert mocked_state.call_args_list == [call()]
    assert mocked_formula.call_args_list == []
    assert mocked_required.call_args_list == [call({})]
    mocker.resetall()
    # update sequences
    filled_application.rules.managed = True
    filled_application.update_sequences()
    sequenced_processes = {process.process_name: process
                           for sub_seq in filled_application.start_sequence.values()
                           for process in sub_seq}
    # test again (check also that failures are reset)
    filled_application.major_failure = True
    filled_application.minor_failure = True
    filled_application.update()
    assert filled_application.state == ApplicationStates.RUNNING
    assert not filled_application.major_failure
    assert not filled_application.minor_failure
    assert mocked_state.call_args_list == [call()]
    assert mocked_formula.call_args_list == []
    assert mocked_required.call_args_list == [call(sequenced_processes)]
    mocker.resetall()
    # test with formula set in application rules
    filled_application.rules.status_formula = "xlogo"
    filled_application.update()
    assert filled_application.state == ApplicationStates.RUNNING
    assert not filled_application.major_failure
    assert not filled_application.minor_failure
    assert mocked_state.call_args_list == [call()]
    assert mocked_formula.call_args_list == [call(sequenced_processes)]
    assert mocked_required.call_args_list == []


def test_application_update_state(filled_application):
    """ Test the rules to update the status of the application method. """
    assert filled_application.state == ApplicationStates.STOPPED
    # there is a process in STOPPING state in the process database
    # STOPPING has the highest priority in application state evaluation
    assert filled_application.update_state() == ApplicationStates.STOPPING
    # set STOPPING process to STOPPED
    for process in filled_application.processes.values():
        if process.state == ProcessStates.STOPPING:
            process.state = ProcessStates.STOPPED
    # now STARTING is expected as it is the second priority
    assert filled_application.update_state() == ApplicationStates.STARTING
    # set STARTING process to RUNNING
    starting_process = next((process for process in filled_application.processes.values()
                             if process.state == ProcessStates.STARTING), None)
    starting_process.state = ProcessStates.RUNNING
    # update status. there is still one BACKOFF process leading to STARTING application
    assert filled_application.update_state() == ApplicationStates.STARTING
    # set BACKOFF process to EXITED unexpected
    backoff_process = next((process for process in filled_application.processes.values()
                            if process.state == ProcessStates.BACKOFF), None)
    backoff_process.state = ProcessStates.EXITED
    backoff_process.expected_exit = False
    # update status. now there is only stopped and running processes.
    assert filled_application.update_state() == ApplicationStates.RUNNING
    # set all process to RUNNING
    for process in filled_application.processes.values():
        process.state = ProcessStates.RUNNING
    assert filled_application.update_state() == ApplicationStates.RUNNING
    # set all processes to STOPPED
    for process in filled_application.processes.values():
        process.state = ProcessStates.STOPPED
    # all processes are STOPPED in a STOPPED application, so no failure
    assert filled_application.update_state() == ApplicationStates.STOPPED


def test_application_update_status_required(filled_application):
    """ Test the rules to update the status of the application method. """
    # initial status
    assert filled_application.state == ApplicationStates.STOPPED
    assert not filled_application.major_failure
    assert not filled_application.minor_failure
    # first call with no process required and no start sequence
    filled_application.update_status_required({})
    assert not filled_application.major_failure
    assert not filled_application.minor_failure
    # set application to managed, update sequences and state
    filled_application.rules.managed = True
    filled_application.update_sequences()
    filled_application.state = filled_application.update_state()
    sequenced_processes = {process.process_name: process
                           for sub_seq in filled_application.start_sequence.values()
                           for process in sub_seq}
    # there is a process in STOPPING state in the process database
    # STOPPING has the highest priority in application state evaluation
    assert filled_application.state == ApplicationStates.STOPPING
    # check the operational status
    filled_application.update_status_required(sequenced_processes)
    # multiple STOPPED processes and global state not STOPPED
    # in default rules, no process is required so this is minor
    assert not filled_application.major_failure
    assert filled_application.minor_failure
    filled_application.minor_failure = False
    # set FATAL process to major
    fatal_process = next((process for process in filled_application.processes.values()
                          if process.state == ProcessStates.FATAL))
    fatal_process.rules.required = True
    # update status. major failure is now expected
    # minor still expected
    filled_application.update_status_required(sequenced_processes)
    assert filled_application.major_failure
    assert not filled_application.minor_failure
    filled_application.major_failure = False
    # set STOPPING process to STOPPED and check no impact on failures
    for process in filled_application.processes.values():
        if process.state == ProcessStates.STOPPING:
            process.state = ProcessStates.STOPPED
    filled_application.update_status_required(sequenced_processes)
    assert filled_application.major_failure
    assert not filled_application.minor_failure
    # set STARTING process to RUNNING and check no impact on failures
    starting_process = next((process for process in filled_application.processes.values()
                             if process.state == ProcessStates.STARTING), None)
    starting_process.state = ProcessStates.RUNNING
    filled_application.update_status_required(sequenced_processes)
    assert filled_application.major_failure
    assert not filled_application.minor_failure
    filled_application.major_failure = False
    # set BACKOFF process to EXITED unexpected
    backoff_process = next((process for process in filled_application.processes.values()
                            if process.state == ProcessStates.BACKOFF), None)
    backoff_process.state = ProcessStates.EXITED
    backoff_process.expected_exit = False
    # update status. major failure is set, so not minor_failure
    filled_application.update_status_required(sequenced_processes)
    assert filled_application.major_failure
    assert not filled_application.minor_failure
    filled_application.major_failure = False
    # set all process to RUNNING
    for process in filled_application.processes.values():
        process.state = ProcessStates.RUNNING
    # no more failures
    filled_application.update_status_required(sequenced_processes)
    assert not filled_application.major_failure
    assert not filled_application.minor_failure
    # set all processes to STOPPED
    for process in filled_application.processes.values():
        process.state = ProcessStates.STOPPED
    # all processes are STOPPED in a STOPPED application, so no failure
    filled_application.state = ApplicationStates.STOPPED
    filled_application.update_status_required(sequenced_processes)
    assert not filled_application.major_failure
    assert not filled_application.minor_failure


def test_application_get_process_status(filled_application):
    """ Test the ApplicationStatus._get_process_status method. """
    assert not filled_application._get_process_status('xlogo')
    assert not filled_application._get_process_status('sleep')
    assert filled_application._get_process_status('segv')
    assert filled_application._get_process_status('xfontsel')
    # force a state to STARTING
    filled_application.processes['segv'].forced_state = ProcessStates.STARTING
    assert filled_application._get_process_status('segv')
    # force a state to EXITED / expected
    filled_application.processes['xfontsel'].forced_state = ProcessStates.EXITED
    filled_application.processes['xfontsel'].expected_exit = True
    assert filled_application._get_process_status('xfontsel')
    # force a state to EXITED / not expected
    filled_application.processes['xfontsel'].expected_exit = False
    assert not filled_application._get_process_status('xfontsel')


def test_application_get_matches(filled_application):
    """ Test the ApplicationStatus._get_matches method. """
    assert filled_application._get_matches('') == []
    assert filled_application._get_matches('dummy') == []
    assert filled_application._get_matches('xlogo') == ['xlogo']
    assert sorted(filled_application._get_matches('x.*')) == ['xclock', 'xfontsel', 'xlogo']
    assert sorted(filled_application._get_matches(r'yeux_[01]{2}')) == ['yeux_00', 'yeux_01']


class Str:
    def __init__(self, value):
        self.s = value


@pytest.mark.skipif(sys.version_info >= (3, 8), reason="ast.Str is replaced by ast.Constant from Python 3.8")
def test_application_evaluate_strings_deprecated(filled_application):
    """ Test the ApplicationStatus.evaluate method with string expression from Python 3.8.
    ast.Str is obsolete since then and replaced by ast.Constant. """
    # test string leaf error
    expr = ast.Str('a string')
    with pytest.raises(ApplicationStatusParseError):
        filled_application.evaluate(expr)
    # test string leaf with process status running (force ast.Str type because it )
    expr = ast.Str('xfontsel')
    assert filled_application.evaluate(expr) is True
    # test string leaf with process status running
    expr = ast.Str('xlogo')
    assert filled_application.evaluate(expr) is False
    # test pattern with single resolution
    expr = ast.Str('xcl.ck')
    assert filled_application.evaluate(expr) is False
    # test pattern with multiple resolution
    expr = ast.Str('(xlogo|sleep)')
    assert filled_application.evaluate(expr) == [False, False]


@pytest.mark.skipif(sys.version_info < (3, 8), reason="ast.Str is replaced by ast.Constant from Python 3.8")
def test_application_evaluate_strings_38(filled_application):
    """ Test the ApplicationStatus.evaluate method with string expression from Python 3.8.
    ast.Str is obsolete since then and replaced by ast.Constant. """
    # force ast.Str type because it is replaced by an ast.Constant at creation
    ast.Str = Str
    # test string leaf error
    expr = ast.Constant('a string')
    with pytest.raises(ApplicationStatusParseError):
        filled_application.evaluate(expr)
    # test string leaf with process status running (just to hit ast.Str)
    expr = ast.Str('xfontsel')
    assert filled_application.evaluate(expr) is True
    # test string leaf with process status running
    expr = ast.Constant('xlogo')
    assert filled_application.evaluate(expr) is False
    # test pattern with single resolution
    expr = ast.Constant('xcl.ck')
    assert filled_application.evaluate(expr) is False
    # test pattern with multiple resolution
    expr = ast.Constant('(xlogo|sleep)')
    assert filled_application.evaluate(expr) == [False, False]


def test_application_evaluate_functions(filled_application):
    """ Test the ApplicationStatus.evaluate method with functions. """
    filled_application.processes['yeux_00'].expected_exit = False
    # test all function
    tree = ast.parse('all("yeux.*")')
    assert filled_application.evaluate(tree.body[0].value) is False
    tree = ast.parse('all("yeux.*1")')
    assert filled_application.evaluate(tree.body[0].value) is True
    # test any function
    tree = ast.parse('any("yeux.*")')
    assert filled_application.evaluate(tree.body[0].value) is True
    tree = ast.parse('any("xlogo")')
    assert filled_application.evaluate(tree.body[0].value) is False
    # test other function
    tree = ast.parse('map(str, ["xlogo"])')
    with pytest.raises(ApplicationStatusParseError) as exc:
        filled_application.evaluate(tree.body[0].value)
    assert exc.value.message == 'unsupported function=map'


def test_application_evaluate_operators(filled_application):
    """ Test the ApplicationStatus.evaluate method with operators. """
    filled_application.processes['yeux_00'].expected_exit = False
    # test unresolvable bool operator
    tree = ast.parse('"xlogo" and "yeux.*"')
    with pytest.raises(ApplicationStatusParseError) as exc:
        filled_application.evaluate(tree.body[0].value)
    assert exc.value.message == 'cannot apply BoolOp on unresolved expression'
    # test and operator
    tree = ast.parse('"xlogo" and "yeux_01"')
    assert filled_application.evaluate(tree.body[0].value) is False
    tree = ast.parse('"xfontsel" and "yeux_01"')
    assert filled_application.evaluate(tree.body[0].value) is True
    # test and operator
    tree = ast.parse('"xlogo" or "yeux_01"')
    assert filled_application.evaluate(tree.body[0].value) is True
    tree = ast.parse('"xlogo" or "yeux_00"')
    assert filled_application.evaluate(tree.body[0].value) is False
    # test unresolvable unary operator
    tree = ast.parse('not "yeux.*"')
    with pytest.raises(ApplicationStatusParseError) as exc:
        filled_application.evaluate(tree.body[0].value)
    assert exc.value.message == 'cannot apply UnaryOp on unresolved expression'
    # test not operator
    tree = ast.parse('not "yeux.00"')
    assert filled_application.evaluate(tree.body[0].value) is True
    tree = ast.parse('not "yeux.01"')
    assert filled_application.evaluate(tree.body[0].value) is False
    # test error when using other unary operator type
    tree = ast.parse('- "yeux.01"')
    with pytest.raises(ApplicationStatusParseError) as exc:
        filled_application.evaluate(tree.body[0].value)
    assert exc.value.message == 'unsupported UnaryOp=USub'
    # test error when using other operator type
    tree = ast.parse('"yeux.01" + "yeux.01"')
    with pytest.raises(ApplicationStatusParseError) as exc:
        filled_application.evaluate(tree.body[0].value)
    assert exc.value.message == 'unsupported Expr=BinOp'


def test_application_update_status_formula(filled_application):
    """ Test the rules to update the status of the application method. """
    filled_application.processes['yeux_00'].expected_exit = False
    # set application to managed, update sequences and state
    filled_application.rules.managed = True
    filled_application.update_sequences()
    sequenced_processes = {process.process_name: process
                           for sub_seq in filled_application.start_sequence.values()
                           for process in sub_seq}
    # check initial state
    assert not filled_application.major_failure
    assert not filled_application.minor_failure
    # test error handling when the evaluation returns a non-boolean result
    filled_application.rules.status_formula = '"yeux.*"'
    filled_application.update_status_formula(sequenced_processes)
    assert filled_application.major_failure
    assert not filled_application.minor_failure
    filled_application.major_failure = False
    # test error handling when the evaluation raises an exception
    filled_application.rules.status_formula = 'map(str, ["xlogo"])'
    filled_application.update_status_formula(sequenced_processes)
    assert filled_application.major_failure
    assert not filled_application.minor_failure
    filled_application.major_failure = False
    # test evaluation with normal major failure
    filled_application.rules.status_formula = 'all("yeux.*")'
    filled_application.update_status_formula(sequenced_processes)
    assert filled_application.major_failure
    assert not filled_application.minor_failure
    filled_application.major_failure = False
    # test evaluation with no major failure, triggering the minor failure evaluation
    # reason: sleep is FATAL
    filled_application.rules.status_formula = 'any("yeux.*")'
    filled_application.update_status_formula(sequenced_processes)
    assert not filled_application.major_failure
    assert filled_application.minor_failure
    filled_application.minor_failure = False
    # force the FATAL state to STOPPED and the expeected exit and check that minor failure disappears
    filled_application.processes['sleep'].forced_state = ProcessStates.STOPPED
    filled_application.processes['yeux_00'].expected_exit = True
    filled_application.update_status_formula(sequenced_processes)
    assert not filled_application.major_failure
    assert not filled_application.minor_failure



# ======================================================================
# Copyright 2016 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

import re
import time
from datetime import timedelta
from math import sqrt
from typing import Dict, List
from urllib.parse import urlparse

from .ttypes import Payload

# Constants
# TICK period in seconds for internal Supvisors heartbeat
TICK_PERIOD = 5

# special characters used in rules
WILDCARD = '*'
HASHTAG = '#'
ATSIGN = '@'


def simple_localtime(now=None):
    """ Returns the local time as a string, without the date. """
    if now is None:
        now = time.time()
    return time.strftime("%H:%M:%S", time.localtime(now))


def simple_duration(duration: float):
    """ Returns the duration as days / hours / minutes / seconds. """
    return str(timedelta(seconds=int(duration)))


# Keys of information kept from Supervisor
__Payload_Keys = ('name', 'group', 'state', 'statename', 'start', 'stop', 'now', 'pid', 'description', 'spawnerr')


def extract_process_info(info: Payload) -> Payload:
    """ Returns a subset of Supervisor process information. """
    payload = {key: info[key] for key in __Payload_Keys if key in info}
    # expand information with 'expected' (deduced from spawnerr)
    payload['expected'] = not info['spawnerr']
    return payload


# parse the Server URL of Supervisor
class SupervisorServerUrl:
    """ Store and update the environment for RPC interfaces. """

    def __init__(self, env: Dict):
        """ Parse the Supervisor server URL for later modification. """
        self.env: Dict = env.copy()
        # get the possible authentication part
        parsed_url = urlparse(env['SUPERVISOR_SERVER_URL'])
        self.authentication = ''
        if parsed_url.username:
            self.authentication = parsed_url.username
            if parsed_url.password:
                self.authentication += f':{parsed_url.password}'
            self.authentication += '@'

    def update_url(self, hostname: str, port: int):
        """ Update the URL by changing the hostname and the port. """
        self.env['SUPERVISOR_SERVER_URL'] = f'http://{self.authentication}{hostname}:{port}'


# simple functions
def mean(x) -> float:
    return sum(x) / float(len(x))


def srate(x, y) -> float:
    return 100.0 * x / y - 100.0 if y else float('inf')


def stddev(lst, avg) -> float:
    return sqrt(sum((x - avg) ** 2 for x in lst) / len(lst))


# bit manipulation
def get_bit(data, num):
    base, shift = int(num // 8), int(num % 8)
    return (data[base] >> shift) & 0x1


def set_bit(data, num, value):
    base, shift = int(num // 8), int(num % 8)
    if value:
        data[base] |= 0x1 << shift
    else:
        data[base] &= ~(0x1 << shift)


# to display small float values
def get_small_value(value) -> str:
    """ Segregate small values from 0.
    For small negative values, the display is already suitable. """
    if 0 < value < 0.005:
        return f'+0.00'
    return f'{value:.2f}'


# linear regression
def get_linear_regression(xdata, ydata):
    """ Calculate the coefficients of the linear equation corresponding
    to the linear regression of a series of points. """
    try:
        import numpy
        return tuple(numpy.polyfit(xdata, ydata, 1))
    except ImportError:
        # numpy not available
        # try something approximate and simple
        datasize = len(xdata)
        sum_x = float(sum(xdata))
        sum_y = float(sum(ydata))
        sum_xx = float(sum(map(lambda x: x * x, xdata)))
        sum_products = float(sum([xdata[i] * ydata[i]
                                  for i in range(datasize)]))
        a = (sum_products - sum_x * sum_y / datasize) / (sum_xx - (sum_x * sum_x) / datasize)
        b = (sum_y - a * sum_x) / datasize
        return a, b


# get statistics from data
def get_stats(xdata, ydata):
    """ Calculate the following statistics from a series of points:
    - the mean value,
    - the instant rate between the two last values,
    - the coefficients of the linear regression,
    - the standard deviation. """
    rate, a, b, dev = (None,) * 4
    # calculate mean value
    avg = mean(ydata)
    if len(ydata) > 1:
        # calculate instant rate value between last 2 values
        rate = srate(ydata[-1], ydata[-2])
        # calculate slope value from linear regression of values
        a, b = get_linear_regression(xdata, ydata)
        # calculate standard deviation
        dev = stddev(ydata, avg)
    return avg, rate, (a, b), dev


# docstring parsing
SUPERVISOR_PARAM_FORMAT = re.compile(r'^@param\s+(?P<type>[a-z]+)\s+(?P<name>\w+)\s+(?P<desc>.*)$')
SUPERVISOR_RETURN_FORMAT = re.compile(r'^@return\s+(?P<type>[a-z]+)(\s+(?P<name>\w+)(\s+(?P<desc>.*))?)?$')

TYPE_FORMAT = r'(?P<type>[\w\[\], ]+)'
SUPVISORS_PARAM_FORMAT = re.compile(rf'^:param\s+{TYPE_FORMAT}\s+(?P<name>\w+):\s+(?P<desc>.*)$')
SUPVISORS_RETURN_FORMAT = re.compile(r'^:return:\s+(?P<desc>.*)$')
SUPVISORS_RTYPE_FORMAT = re.compile(rf'^:rtype:\s+{TYPE_FORMAT}$')
SUPVISORS_RAISE_FORMAT = re.compile(r'^:raises\s+(?P<exc>\w+):\s+(?P<desc>.*)$')


def parse_docstring(comment: str) -> List:
    """ Extract information from the docstring.
    Return the same structure as supervisor.xmlrpc.gettags. """
    description = [0, None, None, None, []]
    parameters = {}
    returns = None
    raises = {}
    # deal with description first
    current_struct = description
    current_desc = description[4]
    # reading fields
    for idx, line in enumerate(comment.split('\n')):
        stripped_line = line.strip()
        if not stripped_line:
            continue
        match = False
        # deal with parameters
        for fmt in [SUPERVISOR_PARAM_FORMAT, SUPVISORS_PARAM_FORMAT]:
            result = fmt.match(stripped_line)
            if result:
                match = True
                param_name = result.group('name')
                current_struct = parameters[param_name] = [idx, 'param', result.group('type'), param_name,
                                                           [result.group('desc')]]
        # deal with Supervisor return
        result = SUPERVISOR_RETURN_FORMAT.match(stripped_line)
        if result:
            match = True
            # Supervisor does not always provide a return name (e.g. signalProcess)
            name = result.group('name') or ''
            # Supervisor does not always provide a return description (e.g. getPID)
            desc = result.group('desc')
            returns = [idx, 'return', result.group('type'), name, [desc] if desc else []]
            current_struct = returns
        # deal with Supvisors return
        result = SUPVISORS_RETURN_FORMAT.match(stripped_line)
        if result:
            match = True
            if returns:
                returns[4] = [result.group('desc')]
            else:
                returns = [idx, 'return', None, None, [result.group('desc')]]
            current_struct = returns
        result = SUPVISORS_RTYPE_FORMAT.match(stripped_line)
        if result:
            match = True
            if returns:
                returns[2] = result.group('type')
            else:
                returns = [idx, 'return', result.group('type'), None, []]
            current_struct = returns
        # deal with Supvisors raises
        result = SUPVISORS_RAISE_FORMAT.match(stripped_line)
        if result:
            match = True
            exc = result.group('exc')
            current_struct = raises[exc] = [idx, 'raises', exc, None, [result.group('desc')]]
        # deal with description
        if match:
            current_desc = current_struct[4]
        else:
            current_desc.append(stripped_line)
    # return as gettags (exceptions not used)
    description[4] = '\n'.join(description[4])
    returns[4] = '\n'.join(returns[4])
    for param in parameters.values():
        param[4] = '\n'.join(param[4])
    return [tuple(description)] + [tuple(param) for param in parameters.values()] + [tuple(returns)]



# ======================================================================
# Copyright 2023 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

from unittest.mock import call

from supvisors.internal_com.rpchandler import *


def test_all_requests(mocker, supvisors):
    """ Test the requests of the RpcHandler communication. """
    handler = RpcHandler(supvisors)
    assert handler.logger is supvisors.logger
    # mock proxy server
    proxy_server = mocker.patch.object(handler, 'proxy_server')
    # test push CHECK_INSTANCE
    handler.send_check_instance('10.0.0.1')
    expected = '10.0.0.1', (RequestHeaders.CHECK_INSTANCE.value, None)
    assert proxy_server.push_request.call_args_list == [call(*expected)]
    proxy_server.push_request.reset_mock()
    # test push START_PROCESS
    handler.send_start_process('10.0.0.1', 'group:name', 'extra args')
    expected = '10.0.0.1', (RequestHeaders.START_PROCESS.value, ('group:name', 'extra args'))
    assert proxy_server.push_request.call_args_list == [call(*expected)]
    proxy_server.push_request.reset_mock()
    # test push STOP_PROCESS
    handler.send_stop_process('10.0.0.1', 'group:name')
    expected = '10.0.0.1', (RequestHeaders.STOP_PROCESS.value, ('group:name',))
    assert proxy_server.push_request.call_args_list == [call(*expected)]
    proxy_server.push_request.reset_mock()
    # test push RESTART
    handler.send_restart('10.0.0.1')
    expected = '10.0.0.1', (RequestHeaders.RESTART.value, None)
    assert proxy_server.push_request.call_args_list == [call(*expected)]
    proxy_server.push_request.reset_mock()
    # test push SHUTDOWN
    handler.send_shutdown('10.0.0.1')
    expected = '10.0.0.1', (RequestHeaders.SHUTDOWN.value, None)
    assert proxy_server.push_request.call_args_list == [call(*expected)]
    proxy_server.push_request.reset_mock()
    # test push RESTART_SEQUENCE
    handler.send_restart_sequence('10.0.0.1')
    expected = '10.0.0.1', (RequestHeaders.RESTART_SEQUENCE.value, None)
    assert proxy_server.push_request.call_args_list == [call(*expected)]
    proxy_server.push_request.reset_mock()
    # test push RESTART_ALL
    handler.send_restart_all('10.0.0.1')
    expected = '10.0.0.1', (RequestHeaders.RESTART_ALL.value, None)
    assert proxy_server.push_request.call_args_list == [call(*expected)]
    proxy_server.push_request.reset_mock()
    # test push SHUTDOWN_ALL
    handler.send_shutdown_all('10.0.0.1')
    expected = '10.0.0.1', (RequestHeaders.SHUTDOWN_ALL.value, None)
    assert proxy_server.push_request.call_args_list == [call(*expected)]
    proxy_server.push_request.reset_mock()
    # check that the other methods have not been called
    assert not proxy_server.post_discovery.called
    assert not proxy_server.publish.called
    # call stop
    handler.stop()
    assert proxy_server.stop.call_args_list == [call()]


def test_all_publications(mocker, supvisors):
    """ Test the publications of the RpcHandler communication. """
    handler = RpcHandler(supvisors)
    assert handler.logger is supvisors.logger
    # mock proxy server
    proxy_server = mocker.patch.object(handler, 'proxy_server')
    # test push TICK
    handler.send_tick_event({'when': 1234})
    expected = PublicationHeaders.TICK.value, {'when': 1234}
    assert proxy_server.push_publication.call_args_list == [call(expected)]
    proxy_server.push_publication.reset_mock()
    # test push PROCESS
    handler.send_process_state_event({'name': 'dummy', 'state': 'RUNNING'})
    expected = PublicationHeaders.PROCESS.value, {'name': 'dummy', 'state': 'RUNNING'}
    assert proxy_server.push_publication.call_args_list == [call(expected)]
    proxy_server.push_publication.reset_mock()
    # test push PROCESS_ADDED
    handler.send_process_added_event({'name': 'dummy'})
    expected = PublicationHeaders.PROCESS_ADDED.value, {'name': 'dummy'}
    assert proxy_server.push_publication.call_args_list == [call(expected)]
    proxy_server.push_publication.reset_mock()
    # test push PROCESS_REMOVED
    handler.send_process_removed_event({'name': 'dummy'})
    expected = PublicationHeaders.PROCESS_REMOVED.value, {'name': 'dummy'}
    assert proxy_server.push_publication.call_args_list == [call(expected)]
    proxy_server.push_publication.reset_mock()
    # test push PROCESS_DISABILITY
    handler.send_process_disability_event({'name': 'dummy_program'})
    expected = PublicationHeaders.PROCESS_DISABILITY.value, {'name': 'dummy_program'}
    assert proxy_server.push_publication.call_args_list == [call(expected)]
    proxy_server.push_publication.reset_mock()
    # test push HOST_STATISTICS
    handler.send_host_statistics({'cpu': 1.0})
    expected = PublicationHeaders.HOST_STATISTICS.value, {'cpu': 1.0}
    assert proxy_server.push_publication.call_args_list == [call(expected)]
    proxy_server.push_publication.reset_mock()
    # test push PROCESS_STATISTICS
    handler.send_process_statistics({'dummy_process': {'cpu': 1.0}})
    expected = PublicationHeaders.PROCESS_STATISTICS.value, {'dummy_process': {'cpu': 1.0}}
    assert proxy_server.push_publication.call_args_list == [call(expected)]
    proxy_server.push_publication.reset_mock()
    # test push STATE
    handler.send_state_event({'state': 'INIT'})
    expected = PublicationHeaders.STATE.value, {'state': 'INIT'}
    assert proxy_server.push_publication.call_args_list == [call(expected)]
    proxy_server.push_publication.reset_mock()
    # check that the other methods have not been called
    assert not proxy_server.post_discovery.called
    assert not proxy_server.post_request.called
    # call stop
    handler.stop()
    assert proxy_server.stop.call_args_list == [call()]


def test_discovery_event(mocker, supvisors):
    """ Test the discovery events of the RpcHandler communication. """
    handler = RpcHandler(supvisors)
    assert handler.logger is supvisors.logger
    # mock proxy server
    proxy_server = mocker.patch.object(handler, 'proxy_server')
    # test push STATE
    expected = (('10.0.0.1', ('10.0.0.1', 7777)), {'state': 'INIT'})
    handler.push_notification(expected)
    assert proxy_server.push_notification.call_args_list == [call(expected)]
    proxy_server.push_notification.reset_mock()
    # check that the other methods have not been called
    assert not proxy_server.post_request.called
    assert not proxy_server.publish.called
    # call stop
    handler.stop()
    assert proxy_server.stop.call_args_list == [call()]
